import collections
import logging
import os
import yaml

logger = logging.getLogger(__name__)


class ConfigurationError(Exception):
    pass


class ConfigItem(object):
    def __init__(self, required=False, default=None, help='', multiple=False):
        self.help = help
        self.required = required
        self.multiple = multiple
        self.default = self.convert(default)

    def convert(self, item, manager=None):
        if not self.multiple:
            return item
        elif (isinstance(item, collections.Sequence)
              and not isinstance(item, basestring)):
            return item
        else:
            return [item]


class ConfigSection(collections.Mapping):
    def __init__(self, required=True, help='', config_description=None):
        self.config_description = config_description
        self.help = help
        self.required = required
        self.default = None

    def convert(self, item, manager):
        return manager.wrap(item, self.config_description)

    def __len__(self):
        return len(self.config_description)

    def __iter__(self):
        return iter(self.config_description)

    def __getitem__(self, key):
        return self.config_description[key]


class ConfigManager(collections.Mapping):

    @classmethod
    def wrap(cls, conf, config_description):
        if hasattr(conf, 'check_config'):
            wrapped_conf = conf
        else:
            wrapped_conf = cls(conf, config_description)
        return wrapped_conf

    def __init__(self, config_dict, config_description):
        self.config_paths = []
        self._configs = dict()
        self._description = config_description
        self._required = set()
        self._defaults = dict()
        for k, item in self._description.items():
            if item.required:
                self._required.add(k)
            if item.default is not None:
                self._defaults[k] = item.default
        for k, item in config_dict.items():
            if k in self._description:
                self._configs[k] = self._description[k].convert(item, self)
            else:
                self._configs[k] = item
        self._keys = set(self._defaults.keys() + self._configs.keys())

    def __len__(self):
        return len(self._keys)

    def __iter__(self):
        return iter(self._keys)

    def __getitem__(self, key):
        if key in self._configs:
            return self._configs[key]
        if key in self._defaults:
            return self._defaults[key]
        raise KeyError(key)

    def add_config_path(self, *args):
        for path in args:
            if path not in self.config_paths:
                self.config_paths.append(path)

    def check_config(self, prefix=''):
        if prefix:
            prefix = prefix + '/'
        for r in self._required:
            if r not in self:
                msg = "Required Configuration setting %s%s is missing!" % (prefix,r)
                logger.error(msg)
                raise ConfigurationError(msg)
        for k, item in self.items():
            if hasattr(item, 'check_config'):
                item.check_config(prefix="%s%s" % (prefix,k))

    @classmethod
    def _load_yaml_config(cls, config_data, filename="(unknown)"):
        """Load a yaml config file."""

        try:
            config = yaml.safe_load(config_data)
        except yaml.YAMLError as err:
            if hasattr(err, 'problem_mark'):
                mark = err.problem_mark
                errmsg = ("Invalid YAML syntax in Configuration file "
                            "%(file)s at line: %(line)s, column: %(column)s."
                          % dict(file=filename,
                                 line=mark.line + 1,
                                 column=mark.column + 1))
            else:
                errmsg = ("YAML error reading Configuration file "
                            "%(file)s"
                          % dict(file=filename))
            logger.error(errmsg)
            raise

        logger.info("Configuration: %s", config)
        return config

    @classmethod
    def _load_file(cls, filename, paths):
        for path in paths:
            fullpath = os.path.join(path, filename)
            if os.path.isfile(fullpath):
                with open(fullpath, 'r') as cf:
                    logger.debug("Loading configuration file: %s", fullpath)
                    return cf.read()
        msg = "Unable to find file %s in %s" % (filename, str(paths))
        logger.info(msg)
        return None

    @classmethod
    def load_config_file(cls, filename, filetype=None, paths=None):
        if not paths:
            paths = ['.']
        if filetype is None:
            if (filename.lower().endswith('.yaml') or
                filename.lower().endswith('.yml')):
                filetype = 'yaml'
            elif filename.lower().endswith('.json'):
                filetype = 'json'
            elif (filename.lower().endswith('.conf') or
                filename.lower().endswith('.ini')):
                filetype = 'ini'
            else:
                filetype = 'yaml'
        data = cls._load_file(filename, paths)
        if data is None:
            raise ConfigurationError("Cannot find or read config file: %s" % filename)
        try:
            loader = getattr(cls, "_load_%s_config" % filetype)
        except AttributeError:
            raise ConfigurationError("Unknown config file type: %s" % filetype)
        return loader(data, filename=filename)

    def load_file(self, filename, filetype=None):
        return self.load_config_file(filename, filetype, paths=self.config_paths)





"""Stream schema.

Revision ID: 44289d1492e6
Revises: 3ab6d7bf80cd
Create Date: 2014-08-07 07:34:14.721111

"""

# revision identifiers, used by Alembic.
revision = '44289d1492e6'
down_revision = '3ab6d7bf80cd'

from alembic import op
import sqlalchemy as sa
from winchester import models


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table('stream',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('first_event', models.PreciseTimestamp(), nullable=False),
    sa.Column('last_event', models.PreciseTimestamp(), nullable=False),
    sa.Column('expire_timestamp', models.PreciseTimestamp(), nullable=True),
    sa.Column('fire_timestamp', models.PreciseTimestamp(), nullable=True),
    sa.Column('name', sa.String(length=255), nullable=False),
    sa.Column('state', sa.Integer(), nullable=False),
    sa.Column('state_serial_no', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index('ix_stream_expire_timestamp', 'stream', ['expire_timestamp'], unique=False)
    op.create_index('ix_stream_fire_timestamp', 'stream', ['fire_timestamp'], unique=False)
    op.create_index('ix_stream_name', 'stream', ['name'], unique=False)
    op.create_index('ix_stream_state', 'stream', ['state'], unique=False)
    op.create_table('dist_trait',
    sa.Column('stream_id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=100), nullable=False),
    sa.Column('type', sa.Integer(), nullable=True),
    sa.Column('dt_string', sa.String(length=255), nullable=True),
    sa.Column('dt_float', sa.Float(), nullable=True),
    sa.Column('dt_int', sa.Integer(), nullable=True),
    sa.Column('dt_datetime', models.PreciseTimestamp(), nullable=True),
    sa.Column('dt_timerange_begin', models.PreciseTimestamp(), nullable=True),
    sa.Column('dt_timerange_end', models.PreciseTimestamp(), nullable=True),
    sa.ForeignKeyConstraint(['stream_id'], ['stream.id'], ),
    sa.PrimaryKeyConstraint('stream_id', 'name')
    )
    op.create_index('ix_dist_trait_dt_datetime', 'dist_trait', ['dt_datetime'], unique=False)
    op.create_index('ix_dist_trait_dt_float', 'dist_trait', ['dt_float'], unique=False)
    op.create_index('ix_dist_trait_dt_int', 'dist_trait', ['dt_int'], unique=False)
    op.create_index('ix_dist_trait_dt_string', 'dist_trait', ['dt_string'], unique=False)
    op.create_index('ix_dist_trait_dt_timerange_begin', 'dist_trait', ['dt_timerange_begin'], unique=False)
    op.create_index('ix_dist_trait_dt_timerange_end', 'dist_trait', ['dt_timerange_end'], unique=False)
    op.create_table('streamevent',
    sa.Column('stream_id', sa.Integer(), nullable=False),
    sa.Column('event_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['event_id'], ['event.id'], ),
    sa.ForeignKeyConstraint(['stream_id'], ['stream.id'], ),
    sa.PrimaryKeyConstraint('stream_id', 'event_id')
    )
    op.create_index('ix_event_generated', 'event', ['generated'], unique=False)
    op.create_index('ix_event_message_id', 'event', ['message_id'], unique=False)
    op.create_index('ix_event_type_id', 'event', ['event_type_id'], unique=False)
    op.create_index('ix_trait_t_datetime', 'trait', ['t_datetime'], unique=False)
    op.create_index('ix_trait_t_float', 'trait', ['t_float'], unique=False)
    op.create_index('ix_trait_t_int', 'trait', ['t_int'], unique=False)
    op.create_index('ix_trait_t_string', 'trait', ['t_string'], unique=False)
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('ix_trait_t_string', table_name='trait')
    op.drop_index('ix_trait_t_int', table_name='trait')
    op.drop_index('ix_trait_t_float', table_name='trait')
    op.drop_index('ix_trait_t_datetime', table_name='trait')
    op.drop_index('ix_event_type_id', table_name='event')
    op.drop_index('ix_event_message_id', table_name='event')
    op.drop_index('ix_event_generated', table_name='event')
    op.drop_table('streamevent')
    op.drop_index('ix_dist_trait_dt_timerange_end', table_name='dist_trait')
    op.drop_index('ix_dist_trait_dt_timerange_begin', table_name='dist_trait')
    op.drop_index('ix_dist_trait_dt_string', table_name='dist_trait')
    op.drop_index('ix_dist_trait_dt_int', table_name='dist_trait')
    op.drop_index('ix_dist_trait_dt_float', table_name='dist_trait')
    op.drop_index('ix_dist_trait_dt_datetime', table_name='dist_trait')
    op.drop_table('dist_trait')
    op.drop_index('ix_stream_state', table_name='stream')
    op.drop_index('ix_stream_name', table_name='stream')
    op.drop_index('ix_stream_fire_timestamp', table_name='stream')
    op.drop_index('ix_stream_expire_timestamp', table_name='stream')
    op.drop_table('stream')
    ### end Alembic commands ###



# Copyright (c) 2014 Dark Secret Software Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import logging
import six

logger = logging.getLogger(__name__)


class NoOpGroup(object):
    def match(self):
        return True

    def mismatch(self, reason):
        return False

    def check(self, value, reason):
        return value


class Group(object):
    def __init__(self, name):
        self._name = name  # Group name
        self._match = 0
        self._mismatch = 0
        self._reasons = {}

    def match(self):
        self._match += 1
        return True

    def mismatch(self, reason):
        count = self._reasons.get(reason, 0)
        self._reasons[reason] = count + 1
        self._mismatch += 1
        return False

    def check(self, value, reason):
        if value:
            return self.match()
        return self.mismatch(reason)


@six.add_metaclass(abc.ABCMeta)
class BaseDebugger(object):
    @abc.abstractmethod
    def reset(self):
        pass

    @abc.abstractmethod
    def get_group(self, name):
        pass

    @abc.abstractmethod
    def bump_counter(self, name, inc=1):
        pass

    @abc.abstractmethod
    def get_debug_level(self):
        pass


class NoOpDebugger(BaseDebugger):
    def __init__(self, *args, **kwargs):
       self.noop_group = NoOpGroup()

    def reset(self):
        pass

    def get_group(self, name):
        return self.noop_group

    def bump_counter(self, name, inc=1):
        pass

    def get_debug_level(self):
        return 0


class DetailedDebugger(BaseDebugger):
    def __init__(self, name, debug_level):
        super(DetailedDebugger, self).__init__()
        self._name = name
        self._debug_level = debug_level
        self.reset()

    def reset(self):
        # If it's not a match or a mismatch it was a fatal error.
        self._groups = {}
        self._counters = {}

    def get_group(self, name):
        group = self._groups.get(name, Group(name))
        self._groups[name] = group
        return group

    def bump_counter(self, name, inc=1):
        self._counters[name] = self._counters.get(name, 0) + inc

    def get_debug_level(self):
        return self._debug_level


class DebugManager(object):
    def __init__(self):
        self._debuggers = {}

    def get_debugger(self, trigger_def):
        name = "n/a"
        level = 2  # Default these unknowns to full debug.
        if trigger_def is not None:
            name = trigger_def.name
            level = trigger_def.debug_level
        debugger = self._debuggers.get(name)
        if not debugger:
            if level > 0:
                debugger = DetailedDebugger(name, level)
            else:
                debugger = NoOpDebugger()
            self._debuggers[name] = debugger
        return debugger

    def dump_group(self, debugger, group_name):
        group = debugger.get_group(group_name)
        logger.info("%s Criteria: %d checks, %d passed" %
                        (group._name,
                         group._match + group._mismatch, group._match))

        if debugger.get_debug_level() > 1:
            for kv in group._reasons.items():
                logger.info(" - %s = %d" % kv)

    def dump_counters(self, debugger):
        for kv in debugger._counters.items():
            logger.info("Counter \"%s\" = %d" % kv)

    def dump_debuggers(self):
        for debugger in self._debuggers.values():
            if debugger.get_debug_level() == 0:
                continue

            logger.info("---- Trigger Definition: %s ----" % debugger._name)
            for name in debugger._groups.keys():
                self.dump_group(debugger, name)

            self.dump_counters(debugger)
            debugger.reset()
            logger.info("----------------------------")



import logging

from yagi.handler import BaseHandler
import yagi.config

from winchester.trigger_manager import TriggerManager
from winchester.config import ConfigManager
from winchester import time_sync


logger = logging.getLogger(__name__)


with yagi.config.defaults_for('winchester') as default:
    default("config_file", "winchester.yaml")


class WinchesterHandler(BaseHandler):
    CONFIG_SECTION = "winchester"
    AUTO_ACK = True

    def __init__(self, app=None, queue_name=None):
        super(WinchesterHandler, self).__init__(app=app, queue_name=queue_name)
        conf_file = self.config_get("config_file")
        config = ConfigManager.load_config_file(conf_file)
        self.time_sync = time_sync.TimeSync(config, publishes=True)
        self.trigger_manager = TriggerManager(config, time_sync=self.time_sync)

    def handle_messages(self, messages, env):
        for notification in self.iterate_payloads(messages, env):
            tyme = notification['timestamp']
            self.time_sync.publish(tyme)
            self.trigger_manager.add_notification(notification)

    def on_idle(self, num_messages, queue_name):
        self.trigger_manager._log_statistics()



from contextlib import contextmanager
import logging
import sqlalchemy
from sqlalchemy import and_, or_
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm.exc import MultipleResultsFound

from winchester import models
from winchester.config import ConfigManager, ConfigSection, ConfigItem


logger = logging.getLogger(__name__)

ENGINES = dict()
SESSIONMAKERS = dict()


class DuplicateError(models.DBException):
    pass


class LockError(models.DBException):
    pass


class NoSuchEventError(models.DBException):
    pass


class NoSuchStreamError(models.DBException):
    pass


def sessioned(func):
    def with_session(self, *args, **kw):
        if 'session' in kw:
            return func(self, *args, **kw)
        else:
            with self.in_session() as session:
                kw['session'] = session
                retval = func(self, *args, **kw)
            return retval
    return with_session


class DBInterface(object):

    @classmethod
    def config_description(cls):
        return dict(url=ConfigItem(required=True,
                                       help="Connection URL for database."),
                   )

    def __init__(self, config):
        self.config = ConfigManager.wrap(config, self.config_description())
        self.db_url = config['url']
        self.echo_sql = config.get('echo_sql', False)

    @property
    def engine(self):
        global ENGINES
        if self.db_url not in ENGINES:
            engine = sqlalchemy.create_engine(self.db_url, echo=self.echo_sql)
            ENGINES[self.db_url] = engine
        return ENGINES[self.db_url]

    @property
    def sessionmaker(self):
        global SESSIONMAKERS
        if self.db_url not in SESSIONMAKERS:
            maker = sessionmaker(bind=self.engine)
            SESSIONMAKERS[self.db_url] = maker
        return SESSIONMAKERS[self.db_url]

    def close(self):
        if self.db_url in ENGINES:
            del ENGINES[self.db_url]
        if self.db_url in SESSIONMAKERS:
            del SESSIONMAKERS[self.db_url]

    def get_session(self):
        return self.sessionmaker(expire_on_commit=False)

    @contextmanager
    def in_session(self):
        """Provide a session scope around a series of operations."""
        session = self.get_session()
        try:
            yield session
            session.commit()
        except IntegrityError:
            session.rollback()
            raise DuplicateError("Duplicate unique value detected!")
        except:
            session.rollback()
            raise
        finally:
            session.close()

    @sessioned
    def get_event_type(self, description, session=None):
        t = session.query(models.EventType).filter(
                                models.EventType.desc == description).first()
        if t is None:
            t = models.EventType(description)
            session.add(t)
        return t

    @sessioned
    def create_event(self, message_id, event_type, generated, traits,
                     session=None):
        event_type = self.get_event_type(event_type, session=session)
        e = models.Event(message_id, event_type, generated)
        for name in traits:
            e[name] = traits[name]
        session.add(e)

    @sessioned
    def get_event_by_message_id(self, message_id, session=None):
        try:
            e = session.query(models.Event).\
                filter(models.Event.message_id == message_id).one()
        except NoResultFound:
            raise NoSuchEventError(
                            "No event found with message_id %s!" % message_id)
        return e.as_dict

    @sessioned
    def find_events(self, from_datetime=None, to_datetime=None,
                    event_name=None, traits=None, mark=None, limit=None,
                    session=None, count=False):

        order_desc = True

        q = session.query(models.Event)
        if mark is not None:
            if mark.startswith('+'):
                order_desc=False
                mark = mark[1:]
            if mark.startswith('-'):
                order_desc=True
                mark = mark[1:]
            if mark:
                if order_desc:
                    q = q.filter(models.Event.id < int(mark, 16))
                else:
                    q = q.filter(models.Event.id > int(mark, 16))
        if from_datetime is not None:
            q = q.filter(models.Event.generated > from_datetime)
        if to_datetime is not None:
            q = q.filter(models.Event.generated <= to_datetime)
        if event_name is not None:
            event_type = self.get_event_type(event_name,
                                             session=session)
            q = q.filter(models.Event.event_type_id == event_type.id)
        if traits is not None:
            for name, val in traits.items():
                q = q.filter(models.Event.traits.any(and_(
                        models.Trait.name == name,
                        models.Trait.value == val)))

        if count:
            q = q.count()
            return [{"count": q}]

        if order_desc:
            q = q.order_by(models.Event.id.desc())
            mark_fmt = '%x'
        else:
            q = q.order_by(models.Event.id.asc())
            mark_fmt = '+%x'

        if limit is not None:
            q = q.limit(limit)

        event_info = []
        for event in q.all():
            info = event.as_dict
            info['_mark'] = mark_fmt % event.id
            event_info.append(info)
        return event_info

    @sessioned
    def get_stream_by_id(self, stream_id, session=None):
        try:
            s = session.query(models.Stream).\
                filter(models.Stream.id == stream_id).one()
        except NoResultFound:
            raise NoSuchStreamError("No stream found with id %s!" % stream_id)
        return s

    @sessioned
    def create_stream(self, trigger_name, initial_event, dist_traits,
                      expire_expr, session=None):
        first_event_time = initial_event['timestamp']
        s = models.Stream(trigger_name, first_event_time)
        for trait_name in dist_traits:
            s[trait_name] = dist_traits[trait_name]
        session.add(s)
        self.add_event_stream(s, initial_event, expire_expr, session=session)
        return s

    @sessioned
    def stream_has_dist_trait(self, stream_id, name, value=None, session=None):
        q = session.query(models.DistinguishingTrait)
        q = q.filter(models.DistinguishingTrait.stream_id == stream_id)
        q = q.filter(models.DistinguishingTrait.name == name)
        if value is not None:
            q = q.filter(models.DistinguishingTrait.value == value)
        dt = q.first()
        if dt is not None:
            dt = dt.as_dict
        return dt

    @sessioned
    def get_stream_events(self, stream, session=None):
        if stream not in session:
            stream = session.merge(stream)
        return [event.as_dict for event in stream.events]

    @sessioned
    def add_event_stream(self, stream, event, expire_expr, session=None):
        if stream not in session:
            session.add(stream)
        message_id = event['message_id']
        timestamp = event['timestamp']
        if timestamp < stream.first_event:
            stream.first_event = timestamp
        if timestamp > stream.last_event:
            stream.last_event = timestamp
        stream.expire_timestamp = expire_expr(first=stream.first_event,
                                              last=stream.last_event).timestamp
        eq = session.query(models.Event)
        eq = eq.filter(models.Event.message_id == message_id)
        e = eq.one()
        stream.events.append(e)
        return e

    @sessioned
    def get_active_stream(self, name, dist_traits, current_time, session=None):
        q = session.query(models.Stream)
        q = q.filter(models.Stream.name == name)
        q = q.filter(models.Stream.state == int(models.StreamState.active))
        q = q.filter(models.Stream.expire_timestamp > current_time)
        for name, val in dist_traits.items():
            q = q.filter(models.Stream.distinguished_by.any(and_(
                    models.DistinguishingTrait.name == name,
                    models.DistinguishingTrait.value == val)))
        return q.first()

    @sessioned
    def stream_ready_to_fire(self, stream, timestamp, session=None):
        if stream not in session:
            session.add(stream)
        stream.fire_timestamp = timestamp

    @sessioned
    def get_ready_streams(self, batch_size, current_time, expire=False, session=None):
        q = session.query(models.Stream)
        if expire:
            states = (int(models.StreamState.active), int(models.StreamState.retry_expire))
        else:
            states = (int(models.StreamState.active), int(models.StreamState.retry_fire))

        q = q.filter(models.Stream.state.in_(states))
        if expire:
            q = q.filter(models.Stream.expire_timestamp < current_time)
        else:
            q = q.filter(models.Stream.fire_timestamp < current_time)
        q = q.limit(batch_size)
        return q.all()

    def set_stream_state(self, stream, state):
        serial = stream.state_serial_no
        stream_id = stream.id
        #we do this in a separate session, as it needs to be atomic.
        with self.in_session() as session:
            q = session.query(models.Stream)
            q = q.filter(models.Stream.id == stream_id)
            q = q.filter(models.Stream.state_serial_no == serial)
            ct = q.update(dict(state=int(state), state_serial_no=serial + 1))
        if ct != 1:
            raise LockError("Optimistic Lock failed!")
        return self.get_stream_by_id(stream_id)

    def reset_stream(self, stream):
        if stream.state == models.StreamState.error:
            return self.set_stream_state(stream, models.StreamState.retry_fire)
        if stream.state == models.StreamState.expire_error:
            return self.set_stream_state(stream, models.StreamState.retry_expire)
        return stream

    @sessioned
    def find_streams(self, count=False, stream_id=None, state=None,
                     older_than=None, younger_than=None,
                     name=None, distinguishing_traits=None,
                     session=None, include_events=False,
                     limit=None, mark=None):

        order_desc = True

        q = session.query(models.Stream)
        if mark is not None:
            if mark.startswith('+'):
                order_desc=False
                mark = mark[1:]
            if mark.startswith('-'):
                order_desc=True
                mark = mark[1:]
            if mark:
                if order_desc:
                    q = q.filter(models.Stream.id < int(mark, 16))
                else:
                    q = q.filter(models.Stream.id > int(mark, 16))
        if stream_id is not None:
            q = q.filter(models.Stream.id == stream_id)
        if state is not None:
            q = q.filter(models.Stream.state == int(state))
        if older_than is not None:
            q = q.filter(models.Stream.first_event < older_than)
        if younger_than is not None:
            q = q.filter(models.Stream.last_event > younger_than)
        if name is not None:
            q = q.filter(models.Stream.name == name)
        if distinguishing_traits is not None:
            for name, val in distinguishing_traits.items():
                q = q.filter(models.Stream.distinguished_by.any(and_(
                        models.DistinguishingTrait.name == name,
                        models.DistinguishingTrait.value == val)))

        if count:
            q = q.count()
            return [{"count": q}]

        if order_desc:
            q = q.order_by(models.Stream.id.desc())
            mark_fmt = '%x'
        else:
            q = q.order_by(models.Stream.id.asc())
            mark_fmt = '+%x'

        if limit is not None:
            q = q.limit(limit)

        stream_info = []
        for stream in q.all():
            info = stream.as_dict
            info['_mark'] = mark_fmt % stream.id
            if include_events:
                info['events'] = self.get_stream_events(stream, session=session)
            stream_info.append(info)
        return stream_info

    @sessioned
    def purge_stream(self, stream, session=None):
        if stream not in session:
            session.add(stream)
        session.delete(stream)




import unittest2 as unittest

import datetime
import mock

from winchester import time_sync


class TestTimeSyncNoEndpoint(unittest.TestCase):
    def setUp(self):
        super(TestTimeSyncNoEndpoint, self).setUp()
        self.time_sync = time_sync.TimeSync({})

    def test_should_update(self):
        now = datetime.datetime.utcnow()
        self.assertIsNone(self.time_sync.last_update)
        self.assertTrue(self.time_sync._should_update(now))

        short = now + datetime.timedelta(seconds=1)
        lng = now + datetime.timedelta(seconds=60)
        self.time_sync.last_update = now

        self.assertFalse(self.time_sync._should_update(short))
        self.assertTrue(self.time_sync._should_update(lng))

    def test_current_time(self):
        with mock.patch.object(self.time_sync, "_get_now") as w:
            w.return_value = "123"
            self.assertEqual(self.time_sync.current_time(), "123")

    def test_publish(self):
        with mock.patch.object(time_sync.dateutil.parser, "parse") as p:
            self.time_sync.publish("foo")
            self.assertEquals(0, p.call_count)


class BlowUp(Exception):
    pass


class TestTimeSyncEndpointPublisher(unittest.TestCase):
    def setUp(self):
        super(TestTimeSyncEndpointPublisher, self).setUp()
        self.time_sync = time_sync.TimeSync(
                    {"time_sync_endpoint":"example.com"}, publishes=True)

    def test_fetch_good(self):
        with mock.patch.object(time_sync.requests, "get") as r:
            response = mock.MagicMock()
            response.text = "now"
            r.return_value = response
            self.assertEquals("now", self.time_sync._fetch())

    def test_fetch_empty(self):
        with mock.patch.object(time_sync.time, "sleep") as t:
            with mock.patch.object(time_sync.requests, "get") as r:
                response = mock.MagicMock()
                response.text = ""
                r.return_value = response
                t.side_effect = BlowUp
                with self.assertRaises(BlowUp):
                    self.time_sync._fetch()

    def test_fetch_None(self):
        with mock.patch.object(time_sync.time, "sleep") as t:
            with mock.patch.object(time_sync.requests, "get") as r:
                response = mock.MagicMock()
                response.text = "None"
                r.return_value = response
                t.side_effect = BlowUp
                with self.assertRaises(BlowUp):
                    self.time_sync._fetch()

    def test_current_time(self):
        self.time_sync.last_tyme = "now"
        with mock.patch.object(self.time_sync, "_should_update") as u:
            self.assertEquals("now", self.time_sync.current_time())
            self.assertEquals(0, u.call_count)

    def test_publish(self):
        with mock.patch.object(time_sync.dateutil.parser, "parse") as p:
            p.return_value = "datetime object"
            with mock.patch.object(self.time_sync, "_should_update") as u:
                u.return_value = True

                with mock.patch.object(time_sync.requests, "post") as r:
                    r.return_value = ""

                    self.time_sync.publish("string datetime")

                    r.assert_called_once_with("example.com/time",
                                              data="string datetime")

    def test_publish_fails(self):
        with mock.patch.object(time_sync.dateutil.parser, "parse") as p:
            p.return_value = "datetime object"
            with mock.patch.object(self.time_sync, "_should_update") as u:
                u.return_value = True
                with mock.patch.object(time_sync.requests, "post") as r:
                    r.side_effect = BlowUp
                    with mock.patch.object(time_sync.logger, "exception") as e:
                        self.time_sync.publish("string datetime")
                        self.assertEquals(1, e.call_count)


class TestTimeSyncEndpointConsumer(unittest.TestCase):
    def setUp(self):
        super(TestTimeSyncEndpointConsumer, self).setUp()
        self.time_sync = time_sync.TimeSync(
                    {"time_sync_endpoint":"example.com"})

    def test_current_time(self):
         with mock.patch.object(self.time_sync, "_should_update") as u:
             u.return_value = True
             with mock.patch.object(time_sync.dateutil.parser, "parse") as p:
                p.return_value = "datetime object"
                with mock.patch.object(self.time_sync, "_fetch") as r:
                    r.return_value = "string datetime"

                    self.assertEquals(self.time_sync.current_time(),
                                      "datetime object")

    def test_current_time_fails(self):
        self.time_sync.last_tyme = "no change"
        with mock.patch.object(self.time_sync, "_should_update") as u:
            u.return_value = True
            with mock.patch.object(self.time_sync, "_fetch") as r:
                r.side_effect = BlowUp
                with mock.patch.object(time_sync.logger, "exception") as e:
                    self.assertEquals(self.time_sync.current_time(),
                                      "no change")
                    self.assertEquals(1, e.call_count)



import unittest2 as unittest

import mock

from winchester import debugging


class TestDebugManager(unittest.TestCase):

    def setUp(self):
        super(TestDebugManager, self).setUp()
        self.debug_manager = debugging.DebugManager()

    def test_get_debugger_none(self):
        debugger = self.debug_manager.get_debugger(None)
        self.assertEquals("n/a", debugger._name)
        self.assertEquals(2, debugger._debug_level)

    def test_get_debugger_off(self):
        tdef = mock.MagicMock(name="tdef")
        tdef.name = "my_trigger"
        tdef.debug_level = 0
        debugger = self.debug_manager.get_debugger(tdef)
        self.assertTrue(isinstance(debugger, debugging.NoOpDebugger))
        self.assertEquals(debugger,
                          self.debug_manager._debuggers['my_trigger'])

        debugger2 = self.debug_manager.get_debugger(tdef)
        self.assertEquals(debugger, debugger2)

    def test_get_debugger_on(self):
        tdef = mock.MagicMock(name="tdef")
        tdef.name = "my_trigger"
        tdef.debug_level = 1
        debugger = self.debug_manager.get_debugger(tdef)
        self.assertTrue(isinstance(debugger, debugging.DetailedDebugger))
        self.assertEquals(debugger,
                          self.debug_manager._debuggers['my_trigger'])

        debugger2 = self.debug_manager.get_debugger(tdef)
        self.assertEquals(debugger, debugger2)

    def test_dump_group_level1(self):
        debugger = mock.MagicMock(name="debugger")
        debugger.get_debug_level.return_value = 1
        group = mock.MagicMock(name="group")
        group._name = "my_group"
        group._match = 1
        group._mismatch = 2
        debugger.get_group.return_value = group
        with mock.patch.object(debugging, "logger") as log:
            self.debug_manager.dump_group(debugger, "my_group")

            log.info.assert_called_once_with(
                                "my_group Criteria: 3 checks, 1 passed")

    def test_dump_group_level2(self):
        debugger = mock.MagicMock(name="debugger")
        debugger.get_debug_level.return_value = 2
        group = mock.MagicMock(name="group")
        group._name = "my_group"
        group._match = 1
        group._mismatch = 2
        group._reasons = {"foo": 12}
        debugger.get_group.return_value = group
        with mock.patch.object(debugging, "logger") as log:
            self.debug_manager.dump_group(debugger, "my_group")

            self.assertEquals(log.info.call_args_list,
                                [mock.call("my_group Criteria: 3 checks, 1 passed"),
                                 mock.call(" - foo = 12")])

    def test_dump_counters(self):
        debugger = mock.MagicMock(name="debugger")
        debugger._counters = {'foo': 12}
        with mock.patch.object(debugging, "logger") as log:
            self.debug_manager.dump_counters(debugger)
            log.info.assert_called_once_with('Counter "foo" = 12')

    def test_dump_debuggers_off(self):
        debugger = mock.MagicMock(name="debugger")
        debugger.get_debug_level.return_value = 0
        self.debug_manager._debuggers = {"foo": debugger}
        with mock.patch.object(debugging, "logger") as log:
            self.debug_manager.dump_debuggers()
            self.assertEqual(0, log.info.call_count)

    def test_dump_debuggers_on(self):
        debugger = mock.MagicMock(name="debugger")
        debugger.get_debug_level.return_value = 1
        debugger._name = "my_debugger"
        group = mock.MagicMock(name="group")
        debugger._groups = {"my_group": group}
        self.debug_manager._debuggers = {"foo": debugger}
        with mock.patch.object(self.debug_manager, "dump_counters") as ctr:
            with mock.patch.object(self.debug_manager, "dump_group") as grp:
                with mock.patch.object(debugging, "logger") as log:
                    self.debug_manager.dump_debuggers()
                    self.assertEquals(log.info.call_args_list,
                      [mock.call("---- Trigger Definition: my_debugger ----"),
                       mock.call("----------------------------")])
                grp.assert_called_once_with(debugger, "my_group")
            ctr.assert_called_once_with(debugger)
        debugger.reset.assert_called_once_with()


class TestDetailedDebugger(unittest.TestCase):

    def setUp(self):
        super(TestDetailedDebugger, self).setUp()
        self.debugger = debugging.DetailedDebugger("my_debugger", 2)

    def test_constructor(self):
        with mock.patch("winchester.debugging.DetailedDebugger.reset") \
                                                                as reset:
            d = debugging.DetailedDebugger("my_debugger", 2)
            reset.assert_called_once_with()

        self.assertEquals(self.debugger._name, "my_debugger")
        self.assertEquals(self.debugger._debug_level, 2)

    def test_reset(self):
        self.assertEquals(self.debugger._groups, {})
        self.assertEquals(self.debugger._counters, {})

    def test_get_group(self):
        self.assertEquals(self.debugger._groups, {})
        g = self.debugger.get_group("foo")
        self.assertEquals(g._name, "foo")
        self.assertTrue(self.debugger._groups['foo'])

    def test_bump_counter(self):
        self.assertEquals(self.debugger._counters, {})
        self.debugger.bump_counter("foo")
        self.assertEquals(self.debugger._counters['foo'], 1)

        self.debugger.bump_counter("foo", 2)
        self.assertEquals(self.debugger._counters['foo'], 3)

    def test_get_debug_level(self):
        self.assertEquals(self.debugger.get_debug_level(), 2)


class TestNoOpDebugger(unittest.TestCase):
    def setUp(self):
        super(TestNoOpDebugger, self).setUp()
        self.debugger = debugging.NoOpDebugger("my_debugger", 2)

    def test_reset(self):
        self.debugger.reset()

    def test_get_group(self):
        g = self.debugger.get_group("foo")
        self.assertEquals(g, self.debugger.noop_group)

    def test_bump_counter(self):
        self.debugger.bump_counter("foo")
        self.debugger.bump_counter("foo", 2)

    def test_get_debug_level(self):
        self.assertEquals(self.debugger.get_debug_level(), 0)


class TestGroup(unittest.TestCase):
    def setUp(self):
        super(TestGroup, self).setUp()
        self.group = debugging.Group("my_group")

    def test_constructor(self):
        self.assertEquals("my_group", self.group._name)
        self.assertEquals(0, self.group._match)
        self.assertEquals(0, self.group._mismatch)
        self.assertEquals({}, self.group._reasons)

    def test_match(self):
        self.assertTrue(self.group.match())
        self.assertEquals(1, self.group._match)

    def test_mismatch(self):
        self.assertFalse(self.group.mismatch("reason"))
        self.assertEquals(1, self.group._mismatch)
        self.assertEquals(1, self.group._reasons['reason'])

    def test_check(self):
        self.assertTrue(self.group.check(True, "reason"))
        self.assertEquals(1, self.group._match)
        self.assertEquals(0, self.group._mismatch)
        self.assertEquals({}, self.group._reasons)

        self.assertTrue(self.group.check(True, "reason"))
        self.assertEquals(2, self.group._match)
        self.assertEquals(0, self.group._mismatch)
        self.assertEquals({}, self.group._reasons)

        self.assertFalse(self.group.check(False, "reason"))
        self.assertEquals(2, self.group._match)
        self.assertEquals(1, self.group._mismatch)
        self.assertEquals(1, self.group._reasons['reason'])

        self.assertFalse(self.group.check(False, "reason"))
        self.assertEquals(2, self.group._match)
        self.assertEquals(2, self.group._mismatch)
        self.assertEquals(2, self.group._reasons['reason'])


class TestNoOpGroup(unittest.TestCase):
    def setUp(self):
        super(TestNoOpGroup, self).setUp()
        self.group = debugging.NoOpGroup()

    def test_match(self):
        self.assertTrue(self.group.match())

    def test_mismatch(self):
        self.assertFalse(self.group.mismatch("reason"))

    def test_check(self):
        self.assertTrue(self.group.check(True, "reason"))
        self.assertFalse(self.group.check(False, "reason"))



import datetime
import dateutil.parser
import logging
import time

import requests

logger = logging.getLogger(__name__)


class TimeSync(object):
    def __init__(self, config={}, publishes=False):
        url = config.get('time_sync_endpoint')
        self.endpoint = None
        if url:
            self.endpoint = "%s/time" % url
        logger.debug("Time sync endpoint=%s" % self.endpoint)
        self.last_update = None
        self.last_tyme = self._get_now()
        self.publishes = publishes

    def _get_now(self):
        # Broken out for testing
        return datetime.datetime.utcnow()

    def _should_update(self, now):
        return (not self.last_update or (now - self.last_update).seconds > 20)

    def _fetch(self):
        while True:
            tyme = requests.get(self.endpoint).text
            if tyme and tyme != "None":
                return tyme
            logger.debug("No time published yet. Waiting ...")
            time.sleep(1)

    def current_time(self):
        now = self._get_now()
        if not self.endpoint:
            return now

        if not self.publishes and self._should_update(now):
            try:
                tyme = self._fetch()
                logger.debug("Requested time, got '%s'" % tyme)
                self.last_tyme = dateutil.parser.parse(tyme)
            except Exception as e:
                logger.exception("Could not get time: %s" % e)
            self.last_update = now

        return self.last_tyme

    def publish(self, tyme):
        if not self.endpoint:
            return

        daittyme = dateutil.parser.parse(tyme)
        self.last_tyme = daittyme
        if self._should_update(daittyme):
            self.last_update = daittyme
            try:
                requests.post(self.endpoint, data=tyme)
                logger.debug("Published time: %s" % tyme)
            except Exception as e:
                logger.exception("Could not publish time: %s" % e)



from alembic import util, command, config
import argparse
import inspect


class AlembicCommandLine(object):
    prog = None
    description = None
    allowed_commands = None

    def __init__(self, prog=None, description=None, allowed_commands=None):
        if prog is not None:
            self.prog = prog
        if description is not None:
            self.description = description
        if allowed_commands is not None:
            self.allowed_commands = allowed_commands


        self.parser = self.generate_options()

    def add_command_options(self, parser, positional, kwargs):
        if 'template' in kwargs:
            parser.add_argument("-t", "--template",
                            default='generic',
                            type=str,
                            help="Setup template for use with 'init'")
        if 'message' in kwargs:
            parser.add_argument("-m", "--message",
                            type=str,
                            help="Message string to use with 'revision'")
        if 'sql' in kwargs:
            parser.add_argument("--sql",
                            action="store_true",
                            help="Don't emit SQL to database - dump to "
                                    "standard output/file instead")
        if 'tag' in kwargs:
            parser.add_argument("--tag",
                            type=str,
                            help="Arbitrary 'tag' name - can be used by "
                            "custom env.py scripts.")
        if 'autogenerate' in kwargs:
            parser.add_argument("--autogenerate",
                            action="store_true",
                            help="Populate revision script with candidate "
                                "migration operations, based on comparison "
                                "of database to model.")
        # "current" command
        if 'head_only' in kwargs:
            parser.add_argument("--head-only",
                                action="store_true",
                                help="Only show current version and "
                                "whether or not this is the head revision.")

        if 'rev_range' in kwargs:
            parser.add_argument("-r", "--rev-range",
                                action="store",
                                help="Specify a revision range; "
                                "format is [start]:[end]")


        positional_help = {
            'directory': "location of scripts directory",
            'revision': "revision identifier"
        }
        for arg in positional:
            parser.add_argument(arg, help=positional_help.get(arg))

    def add_options(self, parser):
        parser.add_argument("-c", "--config",
                            type=str,
                            default="alembic.ini",
                            help="Alternate config file")
        parser.add_argument("-n", "--name",
                            type=str,
                            default="alembic",
                            help="Name of section in .ini file to "
                                    "use for Alembic config")
        parser.add_argument("-x", action="append",
                            help="Additional arguments consumed by "
                            "custom env.py scripts, e.g. -x "
                            "setting1=somesetting -x setting2=somesetting")

    def generate_options(self):
        parser = argparse.ArgumentParser(prog=self.prog)
        self.add_options(parser)
        subparsers = parser.add_subparsers()

        for fn, name, doc, positional, kwarg in self.get_commands():
            subparser = subparsers.add_parser(name, help=doc)
            self.add_command_options(subparser, positional, kwarg)
            subparser.set_defaults(cmd=(fn, positional, kwarg))
        return parser

    def get_commands(self):
        cmds = []
        for fn in [getattr(command, n) for n in dir(command)]:
            if (inspect.isfunction(fn) and
                fn.__name__[0] != '_' and
                fn.__module__ == 'alembic.command'):

                if (self.allowed_commands and
                    fn.__name__ not in self.allowed_commands):
                    continue

                spec = inspect.getargspec(fn)
                if spec[3]:
                    positional = spec[0][1:-len(spec[3])]
                    kwarg = spec[0][-len(spec[3]):]
                else:
                    positional = spec[0][1:]
                    kwarg = []
                cmds.append((fn, fn.__name__, fn.__doc__, positional, kwarg))
        return cmds

    def get_config(self, options):
        return config.Config(file_=options.config,
                             ini_section=options.name,
                             cmd_opts=options)

    def run_cmd(self, config, options):
        fn, positional, kwarg = options.cmd

        try:
            fn(config, *[getattr(options, k) for k in positional],
                        **dict((k, getattr(options, k)) for k in kwarg))
        except util.CommandError as e:
            util.err(str(e))

    def main(self, argv=None):
        options = self.parser.parse_args(argv)
        if not hasattr(options, "cmd"):
            # see http://bugs.python.org/issue9253, argparse
            # behavior changed incompatibly in py3.3
            self.parser.error("too few arguments")
        else:
            self.run_cmd(self.get_config(options), options)


if __name__ == '__main__':
    cmdline = AlembicCommandLine()
    cmdline.main()



import unittest2 as unittest

import mock

import datetime
import timex

from winchester import db as winch_db
from winchester import debugging
from winchester import definition
from winchester import trigger_manager


class TestTriggerManager(unittest.TestCase):

    def setUp(self):
        super(TestTriggerManager, self).setUp()
        self.debugger = debugging.NoOpDebugger()

    @mock.patch.object(trigger_manager.ConfigManager, 'wrap')
    def test_save_event(self, mock_config_wrap):
        tm = trigger_manager.TriggerManager('test')
        tm.db = mock.MagicMock(spec=tm.db)
        event = dict(message_id='1234-test-5678',
                     timestamp=datetime.datetime(2014,8,1,10,9,8,77777),
                     event_type='test.thing',
                     test_trait="foobar",
                     other_test_trait=42)
        self.assertTrue(tm.save_event(event))
        tm.db.create_event.assert_called_once_with('1234-test-5678', 'test.thing',
            datetime.datetime(2014,8,1,10,9,8,77777),
            dict(test_trait='foobar', other_test_trait=42))

    @mock.patch.object(trigger_manager.ConfigManager, 'wrap')
    def test_save_event_dup(self, mock_config_wrap):
        tm = trigger_manager.TriggerManager('test')
        tm.db = mock.MagicMock(spec=tm.db)
        tm.db.create_event.side_effect = winch_db.DuplicateError("test boom!")
        event = dict(message_id='1234-test-5678',
                     timestamp=datetime.datetime(2014,8,1,10,9,8,77777),
                     event_type='test.thing',
                     test_trait="foobar",
                     other_test_trait=42)
        self.assertFalse(tm.save_event(event))
        tm.db.create_event.assert_called_once_with('1234-test-5678', 'test.thing',
            datetime.datetime(2014,8,1,10,9,8,77777),
            dict(test_trait='foobar', other_test_trait=42))

    @mock.patch('winchester.trigger_manager.EventCondenser', autospec=True)
    @mock.patch.object(trigger_manager.ConfigManager, 'wrap')
    def test_convert_notification(self, mock_config_wrap, mock_condenser):
        tm = trigger_manager.TriggerManager('test')
        tm.db = mock.MagicMock(spec=tm.db)
        tm.distiller = mock.MagicMock(spec=tm.distiller)
        test_event = "I'm a test event!"
        tm.distiller.to_event.return_value = True
        cond = mock_condenser.return_value
        cond.validate.return_value = True
        cond.get_event.return_value = test_event
        tm.save_event = mock.MagicMock()
        tm.save_event.return_value = True

        res = tm.convert_notification('test notification here')
        mock_condenser.assert_called_once_with(tm.db)
        cond.clear.assert_called_once_with()
        cond.validate.assert_called_once_with()
        tm.distiller.to_event.assert_called_once_with('test notification here', cond)
        self.assertEquals(res, test_event)

    @mock.patch('winchester.trigger_manager.EventCondenser', autospec=True)
    @mock.patch.object(trigger_manager.ConfigManager, 'wrap')
    def test_convert_notification_dropped(self, mock_config_wrap, mock_condenser):
        tm = trigger_manager.TriggerManager('test')
        tm.db = mock.MagicMock(spec=tm.db)
        tm.distiller = mock.MagicMock(spec=tm.distiller)
        test_event = "I'm a test event!"
        tm.distiller.to_event.return_value = False
        cond = mock_condenser.return_value
        cond.validate.return_value = True
        cond.get_event.return_value = test_event
        tm.save_event = mock.MagicMock()
        tm.save_event.return_value = True

        test_notif = dict(event_type='test.notification.here', message_id='4242-4242')
        res = tm.convert_notification(test_notif)
        mock_condenser.assert_called_once_with(tm.db)
        cond.clear.assert_called_once_with()
        self.assertFalse(cond.validate.called)
        tm.distiller.to_event.assert_called_once_with(test_notif, cond)
        self.assertFalse(tm.save_event.called)
        self.assertIsNone(res)

    @mock.patch('winchester.trigger_manager.EventCondenser', autospec=True)
    @mock.patch.object(trigger_manager.ConfigManager, 'wrap')
    def test_convert_notification_invalid(self, mock_config_wrap, mock_condenser):
        tm = trigger_manager.TriggerManager('test')
        tm.db = mock.MagicMock(spec=tm.db)
        tm.distiller = mock.MagicMock(spec=tm.distiller)
        test_event = "I'm a test event!"
        tm.distiller.to_event.return_value = True
        cond = mock_condenser.return_value
        cond.validate.return_value = False
        cond.get_event.return_value = test_event
        tm.save_event = mock.MagicMock()
        tm.save_event.return_value = True

        test_notif = dict(event_type='test.notification.here', message_id='4242-4242')
        res = tm.convert_notification(test_notif)
        mock_condenser.assert_called_once_with(tm.db)
        cond.clear.assert_called_once_with()
        cond.validate.assert_called_once_with()
        tm.distiller.to_event.assert_called_once_with(test_notif, cond)
        self.assertFalse(tm.save_event.called)
        self.assertIsNone(res)

    @mock.patch.object(trigger_manager.ConfigManager, 'wrap')
    def test_add_or_create_stream(self, mock_config_wrap):
        tm = trigger_manager.TriggerManager('test')
        tm.db = mock.MagicMock(spec=tm.db)
        tm.db.get_active_stream.return_value = 'Existing Stream'
        tm.current_time = mock.MagicMock()
        trigger_def = mock.MagicMock()
        dist_traits = 'some traits'
        event = "eventful!"

        ret = tm._add_or_create_stream(trigger_def, event, dist_traits)
        tm.db.get_active_stream.assert_called_once_with(trigger_def.name,
                                    dist_traits, tm.current_time.return_value)
        self.assertFalse(tm.db.create_stream.called)
        tm.db.add_event_stream.assert_called_once_with(
                                        tm.db.get_active_stream.return_value,
                                        event, trigger_def.expiration)
        self.assertEqual(ret, tm.db.get_active_stream.return_value)

    @mock.patch.object(trigger_manager.ConfigManager, 'wrap')
    def test_add_or_create_stream_create(self, mock_config_wrap):
        tm = trigger_manager.TriggerManager('test')
        tm.db = mock.MagicMock(spec=tm.db)
        tm.db.get_active_stream.return_value = None
        tm.current_time = mock.MagicMock()
        trigger_def = mock.MagicMock()
        dist_traits = 'some traits'
        event = "eventful!"

        ret = tm._add_or_create_stream(trigger_def, event, dist_traits)
        tm.db.get_active_stream.assert_called_once_with(trigger_def.name, dist_traits,
                                                        tm.current_time.return_value)
        tm.db.create_stream.assert_called_once_with(trigger_def.name, event, dist_traits,
                                                    trigger_def.expiration)
        self.assertFalse(tm.db.add_event_stream.called)
        self.assertEqual(ret, tm.db.create_stream.return_value)

    @mock.patch.object(trigger_manager.ConfigManager, 'wrap')
    def test_ready_to_fire(self, mock_config_wrap):
        tm = trigger_manager.TriggerManager('test')
        tm.db = mock.MagicMock(spec=tm.db)
        tm.current_time = mock.MagicMock()
        trigger_def = mock.MagicMock()
        test_stream = mock.MagicMock()

        tm._ready_to_fire(test_stream, trigger_def)
        trigger_def.get_fire_timestamp.assert_called_once_with(tm.current_time.return_value)
        tm.db.stream_ready_to_fire.assert_called_once_with(test_stream,
            trigger_def.get_fire_timestamp.return_value)

    @mock.patch.object(trigger_manager.ConfigManager, 'wrap')
    def test_add_notification(self, mock_config_wrap):
        tm = trigger_manager.TriggerManager('test')
        tm.convert_notification = mock.MagicMock()
        tm.add_event = mock.MagicMock()

        tm.add_notification("test notification")
        tm.convert_notification.assert_called_once_with("test notification")
        tm.add_event.assert_called_once_with(tm.convert_notification.return_value)

    @mock.patch.object(trigger_manager.ConfigManager, 'wrap')
    def test_add_notification_invalid_or_dropped(self, mock_config_wrap):
        tm = trigger_manager.TriggerManager('test')
        tm.convert_notification = mock.MagicMock()
        tm.add_event = mock.MagicMock()
        tm.convert_notification.return_value = None

        tm.add_notification("test notification")
        tm.convert_notification.assert_called_once_with("test notification")
        self.assertFalse(tm.add_event.called)

    @mock.patch.object(trigger_manager.ConfigManager, 'wrap')
    def test_add_event(self, mock_config_wrap):
        tm = trigger_manager.TriggerManager('test')
        tm.db = mock.MagicMock(spec=tm.db)
        tm.trigger_definitions = [mock.MagicMock() for n in range(3)]
        for d in tm.trigger_definitions:
            d.debugger = self.debugger
        m_def = tm.trigger_definitions[2]
        tm.trigger_definitions[0].match.return_value = None
        tm.trigger_definitions[1].match.return_value = None
        event = mock.MagicMock(name='event', spec=dict)
        tm.save_event = mock.MagicMock()
        tm._add_or_create_stream = mock.MagicMock()
        tm._add_or_create_stream.return_value.fire_timestamp = None
        tm._ready_to_fire = mock.MagicMock()
        m_def.should_fire.return_value = True

        tm.add_event(event)
        tm.save_event.assert_called_once_with(event)
        for td in tm.trigger_definitions:
            td.match.assert_called_once_with(event)
        m_def.get_distinguishing_traits.assert_called_once_with(event,
                                                    m_def.match.return_value)
        tm._add_or_create_stream.assert_called_once_with(m_def, event,
            m_def.get_distinguishing_traits.return_value)
        tm.db.get_stream_events.assert_called_once_with(
                                        tm._add_or_create_stream.return_value)
        m_def.should_fire.assert_called_once_with(
                                        tm.db.get_stream_events.return_value)
        tm._ready_to_fire.assert_called_once_with(
                                tm._add_or_create_stream.return_value, m_def)

    @mock.patch.object(trigger_manager.ConfigManager, 'wrap')
    def test_add_event_on_ready_stream(self, mock_config_wrap):
        tm = trigger_manager.TriggerManager('test')
        tm.db = mock.MagicMock(spec=tm.db)
        tm.trigger_definitions = [mock.MagicMock() for n in range(3)]
        m_def = tm.trigger_definitions[2]
        tm.trigger_definitions[0].match.return_value = None
        tm.trigger_definitions[1].match.return_value = None
        event = mock.MagicMock(name='event', spec=dict)
        tm.save_event = mock.MagicMock()
        tm._add_or_create_stream = mock.MagicMock()
        tm._add_or_create_stream.return_value.fire_timestamp = "Fire!"
        tm._ready_to_fire = mock.MagicMock()
        m_def.should_fire.return_value = True

        tm.add_event(event)
        tm.save_event.assert_called_once_with(event)
        for td in tm.trigger_definitions:
            td.match.assert_called_once_with(event)
        m_def.get_distinguishing_traits.assert_called_once_with(event,
                                                    m_def.match.return_value)
        tm._add_or_create_stream.assert_called_once_with(m_def, event,
            m_def.get_distinguishing_traits.return_value)
        self.assertFalse(tm.db.get_stream_events.called)
        self.assertFalse(m_def.should_fire.called)
        self.assertFalse(tm._ready_to_fire.called)

    @mock.patch.object(trigger_manager.ConfigManager, 'wrap')
    def test_add_event_no_match(self, mock_config_wrap):
        tm = trigger_manager.TriggerManager('test')
        tm.db = mock.MagicMock(spec=tm.db)
        tm.trigger_definitions = [mock.MagicMock() for n in range(3)]
        tm.trigger_definitions[0].match.return_value = None
        tm.trigger_definitions[1].match.return_value = None
        tm.trigger_definitions[2].match.return_value = None
        event = mock.MagicMock(name='event', spec=dict)
        tm.save_event = mock.MagicMock()
        tm._add_or_create_stream = mock.MagicMock()
        tm._add_or_create_stream.return_value.fire_timestamp = "Fire!"
        tm._ready_to_fire = mock.MagicMock()

        tm.add_event(event)
        tm.save_event.assert_called_once_with(event)
        for td in tm.trigger_definitions:
            td.match.assert_called_once_with(event)
        for td in tm.trigger_definitions:
            self.assertFalse(td.get_distinguishing_traits.called)
            self.assertFalse(td.should_fire.called)
        self.assertFalse(tm._add_or_create_stream.called)
        self.assertFalse(tm.db.get_stream_events.called)
        self.assertFalse(tm._ready_to_fire.called)



#for Python2.6 compatability.
import unittest2 as unittest

import mock

import datetime
import timex

from winchester import debugging
from winchester import definition


class TestCriterion(unittest.TestCase):

    def setUp(self):
        super(TestCriterion, self).setUp()
        self.fake_group = debugging.NoOpGroup()
        self.fake_debugger = debugging.NoOpDebugger()

    def test_basic_criterion(self):
        c = definition.Criterion(3, 'foo')
        self.assertTrue(c.match({'foo': 3}, self.fake_group))
        self.assertFalse(c.match({'foo': 5}, self.fake_group))
        self.assertFalse(c.match({'bar': 5}, self.fake_group))
        self.assertFalse(c.match({'foo': "booga"}, self.fake_group))

    def test_numeric_criterion(self):
        c = definition.NumericCriterion("3", 'foo')
        self.assertTrue(c.match({'foo': 3}, self.fake_group))
        self.assertFalse(c.match({'foo': 5}, self.fake_group))
        self.assertFalse(c.match({'bar': 5}, self.fake_group))
        self.assertFalse(c.match({'foo': "booga"}, self.fake_group))
        c = definition.NumericCriterion("> 3", 'foo')
        self.assertFalse(c.match({'foo': 3}, self.fake_group))
        self.assertTrue(c.match({'foo': 5}, self.fake_group))
        c = definition.NumericCriterion("< 3", 'foo')
        self.assertFalse(c.match({'foo': 3}, self.fake_group))
        self.assertFalse(c.match({'foo': 5}, self.fake_group))
        self.assertTrue(c.match({'foo': 1}, self.fake_group))
        with self.assertRaises(definition.DefinitionError):
            c = definition.NumericCriterion("zazz", "foo")
        with self.assertRaises(definition.DefinitionError):
            c = definition.NumericCriterion("", "foo")

    def test_float_criterion(self):
        c = definition.FloatCriterion("3.14", 'foo')
        self.assertTrue(c.match({'foo': 3.14}, self.fake_group))
        self.assertFalse(c.match({'foo': 5.2}, self.fake_group))
        self.assertFalse(c.match({'bar': 5.2}, self.fake_group))
        self.assertFalse(c.match({'foo': "booga"}, self.fake_group))
        c = definition.FloatCriterion("> 3.14", 'foo')
        self.assertFalse(c.match({'foo': 3.14}, self.fake_group))
        self.assertTrue(c.match({'foo': 5.2}, self.fake_group))
        c = definition.FloatCriterion("< 3.14", 'foo')
        self.assertFalse(c.match({'foo': 3.14}, self.fake_group))
        self.assertFalse(c.match({'foo': 3.5}, self.fake_group))
        self.assertTrue(c.match({'foo': 3.02}, self.fake_group))
        with self.assertRaises(definition.DefinitionError):
            c = definition.FloatCriterion("zazz", "foo")
        with self.assertRaises(definition.DefinitionError):
            c = definition.FloatCriterion("", "foo")

    def test_time_criterion(self):
        c = definition.TimeCriterion("day", "foo")
        e = dict(timestamp=datetime.datetime(2014,8,1,7,52,31,2),
                 foo=datetime.datetime(2014,8,1,1,2,0,0))
        self.assertTrue(c.match(e, self.fake_group))
        e = dict(timestamp=datetime.datetime(2014,8,1,7,52,31,2),
                 foo=datetime.datetime(2014,8,2,1,2,0,0))
        self.assertFalse(c.match(e, self.fake_group))
        e = dict(timestamp=datetime.datetime(2014,8,1,7,52,31,2),
                 bar=datetime.datetime(2014,8,1,1,2,0,0))
        self.assertFalse(c.match(e, self.fake_group))
        e = dict(timestamp=datetime.datetime(2014,8,1,7,52,31,2),
                 message_id='1234-5678',
                 quux=4,
                 foo=datetime.datetime(2014,8,1,1,2,0,0))
        self.assertTrue(c.match(e, self.fake_group))


class TestCriteria(unittest.TestCase):

    def setUp(self):
        super(TestCriteria, self).setUp()
        self.fake_group = debugging.NoOpGroup()
        self.fake_debugger = debugging.NoOpDebugger()

    def test_defaults(self):
        criteria = definition.Criteria({})
        self.assertEqual(len(criteria.included_types), 1)
        self.assertEqual(len(criteria.excluded_types), 0)
        self.assertEqual(criteria.included_types[0], '*')
        self.assertEqual(criteria.number, 1)
        self.assertIsNone(criteria.timestamp)
        self.assertEqual(len(criteria.map_distinguished_by), 0)
        self.assertEqual(len(criteria.traits), 0)

    def test_event_type_configs(self):
        config = dict(event_type="test.foo.bar")
        criteria = definition.Criteria(config)
        self.assertEqual(len(criteria.included_types), 1)
        self.assertEqual(len(criteria.excluded_types), 0)
        self.assertEqual(criteria.included_types[0], 'test.foo.bar')
        config = dict(event_type="!test.foo.bar")
        criteria = definition.Criteria(config)
        self.assertEqual(len(criteria.included_types), 1)
        self.assertEqual(len(criteria.excluded_types), 1)
        self.assertEqual(criteria.included_types[0], '*')
        self.assertEqual(criteria.excluded_types[0], 'test.foo.bar')
        config = dict(event_type=["test.foo.bar", "!test.wakka.wakka"])
        criteria = definition.Criteria(config)
        self.assertEqual(len(criteria.included_types), 1)
        self.assertEqual(len(criteria.excluded_types), 1)
        self.assertEqual(criteria.included_types[0], 'test.foo.bar')
        self.assertEqual(criteria.excluded_types[0], 'test.wakka.wakka')

    def test_match_type(self):
        config = dict(event_type=["test.foo.bar", "!test.wakka.wakka"])
        criteria = definition.Criteria(config)
        self.assertTrue(criteria.match_type('test.foo.bar'))
        self.assertFalse(criteria.match_type('test.wakka.wakka'))
        self.assertFalse(criteria.match_type('test.foo.baz'))
        config = dict(event_type=["test.foo.*", "!test.wakka.*"])
        criteria = definition.Criteria(config)
        self.assertTrue(criteria.match_type('test.foo.bar'))
        self.assertTrue(criteria.match_type('test.foo.baz'))
        self.assertFalse(criteria.match_type('test.wakka.wakka'))

    def test_match_for_type(self):
        config = dict(event_type=["test.foo.*", "!test.wakka.*"])
        criteria = definition.Criteria(config)
        event1 = dict(event_type = "test.foo.zazz")
        event2 = dict(event_type = "test.wakka.zazz")
        event3 = dict(event_type = "test.boingy")
        self.assertTrue(criteria.match(event1, self.fake_group))
        self.assertFalse(criteria.match(event2, self.fake_group))
        self.assertFalse(criteria.match(event3, self.fake_group))

    def test_match_for_timestamp(self):
        config = dict(timestamp='day($launched_at)')
        criteria = definition.Criteria(config)
        event1 = dict(event_type='test.thing',
                      timestamp=datetime.datetime(2014,8,1,17,16,15,14),
                      launched_at=datetime.datetime(2014,8,1,1,2,3,4))
        event2 = dict(event_type='test.thing',
                      timestamp=datetime.datetime(2014,8,2,17,16,15,14),
                      launched_at=datetime.datetime(2014,8,1,1,2,3,4))
        event3 = dict(event_type='test.thing',
                      timestamp=datetime.datetime(2014,8,2,17,16,15,14))
        self.assertTrue(criteria.match(event1, self.fake_group))
        self.assertFalse(criteria.match(event2, self.fake_group))
        self.assertFalse(criteria.match(event3, self.fake_group))

    def test_match_for_traits(self):
        config = dict(traits=dict(some_trait="test",
                                  launched_at={'datetime': "day"},
                                  memory_mb={'int': "> 2048"},
                                  test_weight={'float': "< 4.02"},
                                  other_trait={'string': 'text here'}))
        criteria = definition.Criteria(config)
        event1 = dict(event_type='test.thing',
                      timestamp=datetime.datetime(2014,8,1,17,16,15,14),
                      launched_at=datetime.datetime(2014,8,1,1,2,3,4),
                      some_trait='test',
                      other_trait='text here',
                      memory_mb=4096,
                      test_weight=3.1415)
        event2 = dict(event_type='test.thing',
                      timestamp=datetime.datetime(2014,8,1,17,16,15,14),
                      launched_at=datetime.datetime(2014,8,1,1,2,3,4),
                      some_trait='foo',
                      other_trait='text here',
                      memory_mb=4096,
                      test_weight=3.1415)
        event3 = dict(event_type='test.thing',
                      timestamp=datetime.datetime(2014,8,1,17,16,15,14),
                      launched_at=datetime.datetime(2014,8,1,1,2,3,4),
                      other_trait='text here',
                      memory_mb=4096,
                      test_weight=3.1415)
        event4 = dict(event_type='test.thing',
                      timestamp=datetime.datetime(2014,8,1,17,16,15,14),
                      launched_at=datetime.datetime(2014,8,2,1,2,3,4),
                      some_trait='test',
                      other_trait='text here',
                      memory_mb=4096,
                      test_weight=3.1415)
        event5 = dict(event_type='test.thing',
                      timestamp=datetime.datetime(2014,8,1,17,16,15,14),
                      launched_at=datetime.datetime(2014,8,1,1,2,3,4),
                      some_trait='test',
                      other_trait='text here',
                      memory_mb=1024,
                      test_weight=3.1415)
        event6 = dict(event_type='test.thing',
                      timestamp=datetime.datetime(2014,8,1,17,16,15,14),
                      launched_at=datetime.datetime(2014,8,1,1,2,3,4),
                      some_trait='test',
                      other_trait='text here',
                      memory_mb=4096,
                      test_weight=6.283)
        self.assertTrue(criteria.match(event1, self.fake_group))
        self.assertFalse(criteria.match(event2, self.fake_group))
        self.assertFalse(criteria.match(event3, self.fake_group))
        self.assertFalse(criteria.match(event4, self.fake_group))
        self.assertFalse(criteria.match(event5, self.fake_group))
        self.assertFalse(criteria.match(event6, self.fake_group))


class TestTriggerDefinition(unittest.TestCase):

    def setUp(self):
        super(TestTriggerDefinition, self).setUp()
        self.debug_manager = debugging.DebugManager()

    def test_config_error_check_and_defaults(self):
        with self.assertRaises(definition.DefinitionError):
            definition.TriggerDefinition(dict(), self.debug_manager)
        with self.assertRaises(definition.DefinitionError):
            definition.TriggerDefinition(dict(name='test_trigger'),
                                         self.debug_manager)
        with self.assertRaises(definition.DefinitionError):
            definition.TriggerDefinition(dict(name='test_trigger',
                                             expiration='$last + 1d'),
                                         self.debug_manager)
        with self.assertRaises(definition.DefinitionError):
            definition.TriggerDefinition(dict(name='test_trigger',
                                             expiration='$last + 1d',
                                             fire_pipeline='test_pipeline'),
                                         self.debug_manager)
        with self.assertRaises(definition.DefinitionError):
            definition.TriggerDefinition(
                dict(name='test_trigger',
                     expiration='$last + 1d',
                     fire_pipeline='test_pipeline',
                     fire_criteria=[dict(event_type='test.thing')]),
                self.debug_manager)
        tdef = definition.TriggerDefinition(
                    dict(name='test_trigger',
                         expiration='$last + 1d',
                         fire_pipeline='test_pipeline',
                         fire_criteria=[dict(event_type='test.thing')],
                         match_criteria=[dict(event_type='test.*')]),
                    self.debug_manager)
        self.assertEqual(len(tdef.distinguished_by), 0)
        self.assertEqual(len(tdef.fire_criteria), 1)
        self.assertIsInstance(tdef.fire_criteria[0], definition.Criteria)
        self.assertEqual(len(tdef.match_criteria), 1)
        self.assertIsInstance(tdef.match_criteria[0], definition.Criteria)
        self.assertEqual(tdef.fire_delay, 0)
        self.assertEqual(len(tdef.load_criteria), 0)

    def test_match_for_criteria(self):
        config = dict(name='test_trigger',
                 expiration='$last + 1d',
                 fire_pipeline='test_pipeline',
                 fire_criteria=[dict(event_type='test.thing')],
                 match_criteria=[dict(event_type='test.*')])
        tdef = definition.TriggerDefinition(config, self.debug_manager)
        event1 = dict(event_type='test.thing')
        event2 = dict(event_type='other.thing')
        self.assertTrue(tdef.match(event1))
        self.assertFalse(tdef.match(event2))
        config = dict(name='test_trigger',
                 expiration='$last + 1d',
                 fire_pipeline='test_pipeline',
                 fire_criteria=[dict(event_type='test.thing')],
                 match_criteria=[dict(event_type='test.*'),
                                 dict(event_type='other.*')])
        tdef = definition.TriggerDefinition(config, self.debug_manager)
        self.assertTrue(tdef.match(event1))
        self.assertTrue(tdef.match(event2))

    def test_match_for_distinguished_traits(self):
        config = dict(name='test_trigger',
                 expiration='$last + 1d',
                 distinguished_by=['instance_id'],
                 fire_pipeline='test_pipeline',
                 fire_criteria=[dict(event_type='test.thing')],
                 match_criteria=[dict(event_type='test.*')])
        tdef = definition.TriggerDefinition(config, self.debug_manager)
        event1 = dict(event_type='test.thing', instance_id='foo')
        event2 = dict(event_type='test.thing')
        self.assertTrue(tdef.match(event1))
        self.assertFalse(tdef.match(event2))

    def test_get_distinguished_traits(self):
        config = dict(name='test_trigger',
                 expiration='$last + 1d',
                 distinguished_by=['instance_id'],
                 fire_pipeline='test_pipeline',
                 fire_criteria=[dict(event_type='test.thing')],
                 match_criteria=[dict(event_type='test.*')])
        event1 = dict(event_type='test.thing', instance_id='foo')
        tdef = definition.TriggerDefinition(config, self.debug_manager)
        mcriteria = tdef.match(event1)
        dt = tdef.get_distinguishing_traits(event1, mcriteria)
        self.assertEqual(len(dt), 1)
        self.assertIn('instance_id', dt)
        self.assertEqual(dt['instance_id'], 'foo')

    def test_get_distinguished_traits_with_timeexpression(self):
        config = dict(name='test_trigger',
                 expiration='$last + 1d',
                 distinguished_by=['instance_id', dict(timestamp='day')],
                 fire_pipeline='test_pipeline',
                 fire_criteria=[dict(event_type='test.thing')],
                 match_criteria=[dict(event_type='test.*')])
        event1 = dict(event_type='test.thing', instance_id='foo',
                      timestamp=datetime.datetime(2014,8,1,20,4,23,444))
        tdef = definition.TriggerDefinition(config, self.debug_manager)
        mcriteria = tdef.match(event1)
        dt = tdef.get_distinguishing_traits(event1, mcriteria)
        self.assertEqual(len(dt), 2)
        self.assertIn('instance_id', dt)
        self.assertEqual(dt['instance_id'], 'foo')
        timerange = timex.TimeRange(datetime.datetime(2014,8,1,0,0,0,0),
                                    datetime.datetime(2014,8,2,0,0,0,0))
        self.assertIn('timestamp', dt)
        self.assertIsInstance(dt['timestamp'], timex.TimeRange)
        self.assertEqual(dt['timestamp'].begin, timerange.begin)
        self.assertEqual(dt['timestamp'].end, timerange.end)

    def test_get_distinguished_traits_with_map(self):
        config = dict(name='test_trigger',
                 expiration='$last + 1d',
                 distinguished_by=['instance_id'],
                 fire_pipeline='test_pipeline',
                 fire_criteria=[dict(event_type='test.thing')],
                 match_criteria=[dict(event_type='test.*',
                                      map_distinguished_by=dict(instance_id='other_id'))])
        event1 = dict(event_type='test.thing', instance_id='foo',
                      other_id='bar')
        tdef = definition.TriggerDefinition(config, self.debug_manager)
        mcriteria = tdef.match(event1)
        dt = tdef.get_distinguishing_traits(event1, mcriteria)
        self.assertEqual(len(dt), 1)
        self.assertIn('instance_id', dt)
        self.assertEqual(dt['instance_id'], 'bar')

    def test_get_fire_timestamp(self):
        config = dict(name='test_trigger',
                 expiration='$last + 1d',
                 distinguished_by=['instance_id'],
                 fire_pipeline='test_pipeline',
                 fire_criteria=[dict(event_type='test.thing')],
                 match_criteria=[dict(event_type='test.*')])
        tdef = definition.TriggerDefinition(config, self.debug_manager)
        test_time = datetime.datetime(2014,8,1,20,4,23,444)
        test_time_plus_1hr = datetime.datetime(2014,8,1,21,4,23,444)
        ft = tdef.get_fire_timestamp(test_time)
        self.assertEqual(ft, test_time)
        config = dict(name='test_trigger',
                 expiration='$last + 1d',
                 fire_delay=3600,
                 distinguished_by=['instance_id'],
                 fire_pipeline='test_pipeline',
                 fire_criteria=[dict(event_type='test.thing')],
                 match_criteria=[dict(event_type='test.*')])
        tdef = definition.TriggerDefinition(config, self.debug_manager)
        ft = tdef.get_fire_timestamp(test_time)
        self.assertEqual(ft, test_time_plus_1hr)

    def test_should_fire(self):
        config = dict(name='test_trigger',
                 expiration='$last + 1d',
                 distinguished_by=['instance_id'],
                 fire_pipeline='test_pipeline',
                 fire_criteria=[dict(event_type='test.thing')],
                 match_criteria=[dict(event_type='test.*')])
        tdef = definition.TriggerDefinition(config, self.debug_manager)
        events1 = [ dict(event_type='test.foobar'),
                    dict(event_type='test.thing'),
                    dict(event_type='test.thing')]
        events2 = [ dict(event_type='test.foobar'),
                    dict(event_type='test.thing')]
        events3 = [ dict(event_type='test.foobar'),
                    dict(event_type='test.whatsit')]
        self.assertTrue(tdef.should_fire(events1))
        self.assertTrue(tdef.should_fire(events2))
        self.assertFalse(tdef.should_fire(events3))
        config = dict(name='test_trigger',
                 expiration='$last + 1d',
                 distinguished_by=['instance_id'],
                 fire_pipeline='test_pipeline',
                 fire_criteria=[dict(event_type='test.thing', number=2)],
                 match_criteria=[dict(event_type='test.*')])
        tdef = definition.TriggerDefinition(config, self.debug_manager)
        self.assertTrue(tdef.should_fire(events1))
        self.assertFalse(tdef.should_fire(events2))
        self.assertFalse(tdef.should_fire(events3))



import argparse
import daemon
import logging
from logging.config import fileConfig


logger = logging.getLogger(__name__)


from winchester.config import ConfigManager
from winchester.pipeline_manager import PipelineManager
from winchester import time_sync


def main():
    parser = argparse.ArgumentParser(description="Winchester pipeline worker")
    parser.add_argument('--config', '-c', default='winchester.yaml',
                        help='The name of the winchester config file')
    parser.add_argument('--name', '-n', default='pipeline_worker',
                        help='The name of this process for logging purposes')
    parser.add_argument('--daemon', '-d', help='Run in daemon mode.')
    args = parser.parse_args()

    conf = ConfigManager.load_config_file(args.config)
    proc_name = args.name

    if 'log_level' in conf:
        level = conf['log_level']
        level = getattr(logging, level.upper())
    else:
        level = logging.INFO

    if 'log_file' in conf:
        log_file = conf['log_file'] % dict(proc_name=proc_name)
    else:
        log_file = '%(proc_name)s.log' % dict(proc_name=proc_name)

    # This is a hack, but it's needed to pass the logfile name & default
    # loglevel into log handlers configured with a config file. (mdragon)
    logging.LOCAL_LOG_FILE = log_file
    logging.LOCAL_DEFAULT_LEVEL = level

    if 'logging_config' in conf:
        fileConfig(conf['logging_config'])
    else:
        logging.basicConfig()
        logging.getLogger('winchester').setLevel(level)
    timesync = time_sync.TimeSync(conf)
    pipe = PipelineManager(conf, time_sync=timesync, proc_name=proc_name)
    if args.daemon:
        print "Backgrounding for daemon mode."
        with daemon.DaemonContext():
            pipe.run()
    else:
        pipe.run()


if __name__ == '__main__':
    main()



"""Initial Event schema.

Revision ID: 3ab6d7bf80cd
Revises: None
Create Date: 2014-06-26 01:36:42.792353

"""

# revision identifiers, used by Alembic.
revision = '3ab6d7bf80cd'
down_revision = None

from alembic import op
import sqlalchemy as sa
from winchester import models

def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table('event_type',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('desc', sa.String(length=255), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('desc')
    )
    op.create_table('event',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('message_id', sa.String(length=50), nullable=True),
    sa.Column('generated', models.PreciseTimestamp(), nullable=True),
    sa.Column('event_type_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['event_type_id'], ['event_type.id'], ),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('message_id')
    )
    op.create_table('trait',
    sa.Column('event_id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=100), nullable=False),
    sa.Column('type', sa.Integer(), nullable=True),
    sa.Column('t_string', sa.String(length=255), nullable=True),
    sa.Column('t_float', sa.Float(), nullable=True),
    sa.Column('t_int', sa.Integer(), nullable=True),
    sa.Column('t_datetime', models.PreciseTimestamp(), nullable=True),
    sa.ForeignKeyConstraint(['event_id'], ['event.id'], ),
    sa.PrimaryKeyConstraint('event_id', 'name')
    )
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('trait')
    op.drop_table('event')
    op.drop_table('event_type')
    ### end Alembic commands ###



from datetime import datetime
from decimal import Decimal
import calendar
from enum import IntEnum

import timex

from sqlalchemy import event
from sqlalchemy import and_, or_
from sqlalchemy import literal_column
from sqlalchemy import Column, Table, ForeignKey, Index, UniqueConstraint
from sqlalchemy import Float, Boolean, Text, DateTime, Integer, String
from sqlalchemy import cast, null, case
from sqlalchemy.orm.interfaces import PropComparator
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.dialects.mysql import DECIMAL
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import composite
from sqlalchemy.orm import backref
from sqlalchemy.orm import relationship
from sqlalchemy.orm.collections import attribute_mapped_collection
from sqlalchemy.types import TypeDecorator, DATETIME


class Datatype(IntEnum):
    none = 0
    string = 1
    int = 2
    float = 3
    datetime = 4
    timerange = 5


class StreamState(IntEnum):
    active = 1
    firing = 2
    expiring = 3
    error = 4
    expire_error = 5
    completed = 6
    retry_fire = 7
    retry_expire = 8


class DBException(Exception):
    pass


class InvalidTraitType(DBException):
    pass


def dt_to_decimal(dt):
    t_sec = calendar.timegm(dt.utctimetuple()) + (dt.microsecond/1e6)
    return Decimal("%.6f" % t_sec)


def decimal_to_dt(decimal_timestamp):
    return datetime.utcfromtimestamp(float(decimal_timestamp))


class PreciseTimestamp(TypeDecorator):
    """Represents a timestamp precise to the microsecond."""

    impl = DATETIME

    def load_dialect_impl(self, dialect):
        if dialect.name == 'mysql':
            return dialect.type_descriptor(DECIMAL(precision=20,
                                                   scale=6,
                                                   asdecimal=True))
        return dialect.type_descriptor(DATETIME())

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'mysql':
            return dt_to_decimal(value)
        return value

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'mysql':
            return decimal_to_dt(value)
        return value


class DBTimeRange(object):
    def __init__(self, begin, end):
        self.begin = begin
        self.end = end

    def __composite_values__(self):
        return self.begin, self.end

    def __repr__(self):
        return "DBTimeRange(begin=%r, end=%r)" % (self.begin, self.end)

    def __eq__(self, other):
        return isinstance(other, DBTimeRange) and \
            other.begin == self.begin and \
            other.end == self.end

    def __ne__(self, other):
        return not self.__eq__(other)


class ProxiedDictMixin(object):
    """Adds obj[name] access to a mapped class.

    This class basically proxies dictionary access to an attribute
    called ``_proxied``.  The class which inherits this class
    should have an attribute called ``_proxied`` which points to a dictionary.

    """

    def __len__(self):
        return len(self._proxied)

    def __iter__(self):
        return iter(self._proxied)

    def __getitem__(self, name):
        return self._proxied[name]

    def __contains__(self, name):
        return name in self._proxied

    def __setitem__(self, name, value):
        self._proxied[name] = value

    def __delitem__(self, name):
        del self._proxied[name]


class PolymorphicVerticalProperty(object):
    """A name/value pair with polymorphic value storage."""

    ATTRIBUTE_MAP = {Datatype.none: None}
    PY_TYPE_MAP = {unicode: Datatype.string,
                   int: Datatype.int,
                   float: Datatype.float,
                   datetime: Datatype.datetime,
                   DBTimeRange: Datatype.timerange}

    def __init__(self, name, value=None):
        self.name = name
        self.value = value

    @classmethod
    def get_type_value(cls, value):
        if value is None:
            return Datatype.none, None
        if isinstance(value, str):
            value = value.decode('utf8', 'ignore')
        if isinstance(value, timex.Timestamp):
            value = value.timestamp
        if isinstance(value, timex.TimeRange):
            value = DBTimeRange(value.begin, value.end)
        if type(value) in cls.PY_TYPE_MAP:
            return cls.PY_TYPE_MAP[type(value)], value
        return None, value

    @hybrid_property
    def value(self):
        if self.type not in self.ATTRIBUTE_MAP:
            raise InvalidTraitType("Invalid trait type in db for %s: %s" % (self.name, self.type))
        attribute = self.ATTRIBUTE_MAP[self.type]
        if attribute is None:
            return None
        if self.type == Datatype.timerange:
            val = getattr(self, attribute)
            return timex.TimeRange(val.begin, val.end)
        else:
            return getattr(self, attribute)

    @value.setter
    def value(self, value):
        datatype, value = self.get_type_value(value)
        if datatype not in self.ATTRIBUTE_MAP:
            raise InvalidTraitType("Invalid trait type for %s: %s" % (self.name, datatype))
        attribute = self.ATTRIBUTE_MAP[datatype]
        self.type = int(datatype)
        if attribute is not None:
            setattr(self, attribute, value)

    @value.deleter
    def value(self):
        self._set_value(None)

    @value.comparator
    class value(PropComparator):
        """A comparator for .value, builds a polymorphic comparison.

        """
        def __init__(self, cls):
            self.cls = cls

        def __eq__(self, other):
            dtype, value = self.cls.get_type_value(other)
            if dtype is None:
                dtype = Datatype.string
            if dtype == Datatype.none:
                return self.cls.type == int(Datatype.none)
            attr = getattr(self.cls, self.cls.ATTRIBUTE_MAP[dtype])
            return and_(attr == value, self.cls.type == int(dtype))

        def __ne__(self, other):
            dtype, value = self.cls.get_type_value(other)
            if dtype is None:
                dtype = Datatype.string
            if dtype == Datatype.none:
                return self.cls.type != int(Datatype.none)
            attr = getattr(self.cls, self.cls.ATTRIBUTE_MAP[dtype])
            return and_(attr != value, self.cls.type == int(dtype))

    def __repr__(self):
        return '<%s %r=%r>' % (self.__class__.__name__, self.name, self.value)


Base = declarative_base()


class Trait(PolymorphicVerticalProperty, Base):
    __tablename__ = 'trait'
    __table_args__ = (
        Index('ix_trait_t_int', 't_int'),
        Index('ix_trait_t_string', 't_string'),
        Index('ix_trait_t_datetime', 't_datetime'),
        Index('ix_trait_t_float', 't_float'),
    )
    event_id = Column(Integer, ForeignKey('event.id'), primary_key=True)
    name = Column(String(100), primary_key=True)
    type = Column(Integer)

    ATTRIBUTE_MAP = {Datatype.none: None,
                     Datatype.string: 't_string',
                     Datatype.int: 't_int',
                     Datatype.float: 't_float',
                     Datatype.datetime: 't_datetime',}

    t_string = Column(String(255), nullable=True, default=None)
    t_float = Column(Float, nullable=True, default=None)
    t_int = Column(Integer, nullable=True, default=None)
    t_datetime = Column(PreciseTimestamp(),
                      nullable=True, default=None)

    def __repr__(self):
        return "<Trait(%s) %s=%s/%s/%s/%s on %s>" % (self.name,
                                                     self.type,
                                                     self.t_string,
                                                     self.t_float,
                                                     self.t_int,
                                                     self.t_datetime,
                                                     self.event_id)


class EventType(Base):
    """Types of event records."""
    __tablename__ = 'event_type'

    id = Column(Integer, primary_key=True)
    desc = Column(String(255), unique=True)

    def __init__(self, event_type):
        self.desc = event_type

    def __repr__(self):
        return "<EventType: %s>" % self.desc


class Event(ProxiedDictMixin, Base):
    __tablename__ = 'event'
    __table_args__ = (
        Index('ix_event_message_id', 'message_id'),
        Index('ix_event_type_id', 'event_type_id'),
        Index('ix_event_generated', 'generated')
    )
    id = Column(Integer, primary_key=True)
    message_id = Column(String(50), unique=True)
    generated = Column(PreciseTimestamp())

    event_type_id = Column(Integer, ForeignKey('event_type.id'))
    event_type = relationship("EventType", backref=backref('event_type'))

    traits = relationship("Trait",
                    collection_class=attribute_mapped_collection('name'))
    _proxied = association_proxy("traits", "value",
                            creator=lambda name, value: Trait(name=name, value=value))

    @property
    def event_type_string(self):
        return self.event_type.desc

    @property
    def as_dict(self):
        d = dict(self._proxied)
        d['message_id'] = self.message_id
        d['event_type'] = self.event_type_string
        d['timestamp'] = self.generated
        return d

    def __init__(self, message_id, event_type, generated):

         self.message_id = message_id
         self.event_type = event_type
         self.generated = generated

    def __repr__(self):
        return "<Event %s ('Event : %s %s, Generated: %s')>" % (self.id,
                                                              self.message_id,
                                                              self.event_type,
                                                              self.generated)


stream_event_table = Table('streamevent', Base.metadata,
        Column('stream_id', Integer, ForeignKey('stream.id'), primary_key=True),
        Column('event_id', Integer,
                ForeignKey('event.id'),
                primary_key=True)
        )


class Stream(ProxiedDictMixin, Base):
    __tablename__ = 'stream'

    __table_args__ = (
        Index('ix_stream_name', 'name'),
        Index('ix_stream_state', 'state'),
        Index('ix_stream_expire_timestamp', 'expire_timestamp'),
        Index('ix_stream_fire_timestamp', 'fire_timestamp')
    )
    id = Column(Integer, primary_key=True)
    first_event = Column(PreciseTimestamp(), nullable=False)
    last_event = Column(PreciseTimestamp(), nullable=False)
    expire_timestamp = Column(PreciseTimestamp())
    fire_timestamp = Column(PreciseTimestamp())
    name = Column(String(255), nullable=False)
    state = Column(Integer, default=StreamState.active, nullable=False)
    state_serial_no = Column(Integer, default=0, nullable=False)

    distinguished_by = relationship("DistinguishingTrait",
                    cascade="save-update, merge, delete, delete-orphan",
                    collection_class=attribute_mapped_collection('name'))
    _proxied = association_proxy("distinguished_by", "value",
                    creator=lambda name, value: DistinguishingTrait(name=name, value=value))

    events = relationship(Event, secondary=stream_event_table,
                          order_by=Event.generated)

    @property
    def distinguished_by_dict(self):
        return dict(self._proxied)

    @property
    def as_dict(self):
        return {'name': self.name,
                'id': self.id,
                'state': StreamState(self.state).name,
                'first_event': self.first_event,
                'last_event': self.last_event,
                'fire_timestamp': self.fire_timestamp,
                'expire_timestamp': self.expire_timestamp,
                'distinguishing_traits': self.distinguished_by_dict}



    def __init__(self, name, first_event, last_event=None, expire_timestamp=None,
                fire_timestamp=None, state=None, state_serial_no=None):
        self.name = name
        self.first_event = first_event
        if last_event is None:
            last_event = first_event
        self.last_event = last_event
        self.expire_timestamp = expire_timestamp
        self.fire_timestamp = fire_timestamp
        if state is None:
            state = StreamState.active
        self.state = int(state)
        if state_serial_no is None:
            state_serial_no = 0
        self.state_serial_no = state_serial_no


class DistinguishingTrait(PolymorphicVerticalProperty, Base):
    __tablename__ = 'dist_trait'
    __table_args__ = (
        Index('ix_dist_trait_dt_int', 'dt_int'),
        Index('ix_dist_trait_dt_float', 'dt_float'),
        Index('ix_dist_trait_dt_string', 'dt_string'),
        Index('ix_dist_trait_dt_datetime', 'dt_datetime'),
        Index('ix_dist_trait_dt_timerange_begin', 'dt_timerange_begin'),
        Index('ix_dist_trait_dt_timerange_end', 'dt_timerange_end'),
    )
    stream_id = Column(Integer, ForeignKey('stream.id'), primary_key=True)
    name = Column(String(100), primary_key=True)
    type = Column(Integer)


    ATTRIBUTE_MAP = {Datatype.none:     None,
                     Datatype.string:   'dt_string',
                     Datatype.int:      'dt_int',
                     Datatype.float:    'dt_float',
                     Datatype.datetime: 'dt_datetime',
                     Datatype.timerange:'dt_timerange',
                    }

    dt_string = Column(String(255), nullable=True, default=None)
    dt_float = Column(Float, nullable=True, default=None)
    dt_int = Column(Integer, nullable=True, default=None)
    dt_datetime = Column(PreciseTimestamp(),
                      nullable=True, default=None)
    dt_timerange_begin = Column(PreciseTimestamp(), nullable=True, default=None)
    dt_timerange_end = Column(PreciseTimestamp(), nullable=True, default=None)

    dt_timerange = composite(DBTimeRange, dt_timerange_begin, dt_timerange_end)

    @property
    def as_dict(self):
        return {self.name: self.value}

    def __repr__(self):
        return "<DistinguishingTrait(%s) %s=%s/%s/%s/%s/(%s to %s) on %s>" % (self.name,
                                                             self.type,
                                                             self.dt_string,
                                                             self.dt_float,
                                                             self.dt_int,
                                                             self.dt_datetime,
                                                             self.dt_timerange_begin,
                                                             self.dt_timerange_end,
                                                             self.stream_id)



import unittest2 as unittest

import datetime
import mock

from winchester import pipeline_handler


class TestUsageHandler(unittest.TestCase):
    def setUp(self):
        super(TestUsageHandler, self).setUp()
        self.handler = pipeline_handler.UsageHandler()

    def test_get_audit_period(self):
        event = {}
        apb, ape = self.handler._get_audit_period(event)
        self.assertIsNone(apb)
        self.assertIsNone(ape)

        event = {'audit_period_beginning': "beginning"}
        apb, ape = self.handler._get_audit_period(event)
        self.assertEqual(apb, "beginning")
        self.assertIsNone(ape)

        event = {'audit_period_ending': "ending"}
        apb, ape = self.handler._get_audit_period(event)
        self.assertIsNone(apb)
        self.assertEqual(ape, "ending")

    def test_is_exists(self):
        event = {'event_type': None}
        self.assertFalse(self.handler._is_exists(event))

        event = {'event_type': 'foo'}
        self.assertFalse(self.handler._is_exists(event))

        event = {'event_type': 'compute.instance.exists'}
        self.assertTrue(self.handler._is_exists(event))

    def test_is_non_EOD_exists(self):
        start = datetime.datetime(2014, 12, 31, 0, 0, 0)
        end = start + datetime.timedelta(days=1)
        eod = {'event_type': 'compute.instance.exists',
               'audit_period_beginning': start,
               'audit_period_ending': end}
        self.assertFalse(self.handler._is_non_EOD_exists(eod))

        start = datetime.datetime(2014, 12, 31, 1, 0, 0)
        end = start + datetime.timedelta(hours=1)
        eod = {'event_type': 'compute.instance.exists',
               'audit_period_beginning': start,
               'audit_period_ending': end}
        self.assertTrue(self.handler._is_non_EOD_exists(eod))

        eod = {'event_type': 'compute.instance.foo',
               'audit_period_beginning': start,
               'audit_period_ending': end}
        self.assertFalse(self.handler._is_non_EOD_exists(eod))

        eod = {'event_type': 'compute.instance.foo',
               'audit_period_ending': end}
        self.assertFalse(self.handler._is_non_EOD_exists(eod))

    def test_is_EOD_exists(self):
        start = datetime.datetime(2014, 12, 31, 1, 0, 0)
        end = start + datetime.timedelta(days=1)
        eod = {'event_type': 'compute.instance.exists',
               'audit_period_beginning': start,
               'audit_period_ending': end}
        self.assertFalse(self.handler._is_EOD_exists(eod))

        start = datetime.datetime(2014, 12, 31, 0, 0, 0)
        end = start + datetime.timedelta(hours=1)
        eod = {'event_type': 'compute.instance.exists',
               'audit_period_beginning': start,
               'audit_period_ending': end}
        self.assertFalse(self.handler._is_EOD_exists(eod))

        start = datetime.datetime(2014, 12, 31, 0, 0, 0)
        end = start + datetime.timedelta(days=1)
        eod = {'event_type': 'compute.instance.exists',
               'audit_period_beginning': start,
               'audit_period_ending': end}
        self.assertTrue(self.handler._is_EOD_exists(eod))

        eod = {'event_type': 'compute.instance.foo',
               'audit_period_beginning': start,
               'audit_period_ending': end}
        self.assertFalse(self.handler._is_EOD_exists(eod))

        eod = {'event_type': 'compute.instance.foo',
               'audit_period_ending': end}
        self.assertFalse(self.handler._is_EOD_exists(eod))

    def test_extract_launched_at(self):
        with self.assertRaises(pipeline_handler.UsageException):
            self.handler._extract_launched_at({})
        self.assertEquals("foo", self.handler._extract_launched_at(
                                                    {'launched_at': 'foo'}))

    def test_extract_interesting(self):
        interesting = ["a", "b", "c"]
        e1 = {'event_type': 'a'}
        e2 = {'event_type': 'b'}
        e3 = {'event_type': 'c'}
        e4 = {'event_type': 'd'}
        e5 = {'event_type': 'e'}
        self.assertEquals([e1, e2, e3],
                          self.handler._extract_interesting_events(
                            [e4, e1, e2, e3, e5], interesting))

    def test_verify_fields_no_match(self):
        exists = {'a': 1, 'b': 2, 'c': 3}
        launched = exists
        self.handler._verify_fields(exists, launched, ['d', 'e', 'f'])

    def test_verify_fields_happyday(self):
        exists = {'a': 1, 'b': 2, 'c': 3}
        launched = exists
        self.handler._verify_fields(exists, launched, ['a', 'b', 'c'])

    def test_verify_fields_mismatch(self):
        exists = {'a': 1, 'b': 2, 'c': 3}
        launched = {'a': 10, 'b': 20, 'c': 30}
        with self.assertRaises(pipeline_handler.UsageException):
            self.handler._verify_fields(exists, launched, ['a', 'b', 'c'])

    def test_confirm_delete_no_delete_events(self):
        with self.assertRaises(pipeline_handler.UsageException) as e:
            self.handler._confirm_delete({'deleted_at': 'now',
                                          'state': 'active'}, [], [])
            self.assertEquals("U3", e.code)

        deleted_at = datetime.datetime(2014, 12, 31, 1, 0, 0)
        launched_at = datetime.datetime(2014, 12, 31, 2, 0, 0)
        with self.assertRaises(pipeline_handler.UsageException) as e:
            self.handler._confirm_delete({'deleted_at': deleted_at,
                                          'launched_at': launched_at,
                                          'state': 'deleted'}, [], [])
            self.assertEquals("U4", e.code)

        apb = datetime.datetime(2014, 12, 30, 0, 0, 0)
        ape = datetime.datetime(2014, 12, 31, 0, 0, 0)
        deleted_at = datetime.datetime(2014, 12, 30, 2, 0, 0)
        launched_at = datetime.datetime(2014, 12, 30, 1, 0, 0)
        with self.assertRaises(pipeline_handler.UsageException) as e:
            self.handler._confirm_delete({'deleted_at': deleted_at,
                                          'launched_at': launched_at,
                                          'audit_period_beginning': apb,
                                          'audit_period_ending': ape,
                                          'state': 'deleted'}, [], [])
            self.assertEquals("U5", e.code)

        # Test the do-nothing scenario
        self.handler._confirm_delete({}, [], [])

    def test_confirm_delete_with_delete_events(self):
        with self.assertRaises(pipeline_handler.UsageException) as e:
            self.handler._confirm_delete({}, [{}], [])
            self.assertEquals("U6", e.code)

        with self.assertRaises(pipeline_handler.UsageException) as e:
            self.handler._confirm_delete({'deleted_at': 'now'}, [{}, {}], [])
            self.assertEquals("U7", e.code)

        with mock.patch.object(self.handler, "_verify_fields") as v:
            exists = {'deleted_at': 'now', 'state': 'deleted'}
            deleted = {'foo': 1}
            self.handler._confirm_delete(exists, [deleted], ['a'])
            v.assert_called_with(exists, deleted, ['a'])

    def test_confirm_launched_at(self):
        self.handler._confirm_launched_at([], {'state': 'deleted'})

        apb = datetime.datetime(2014, 12, 30, 0, 0, 0)
        ape = datetime.datetime(2014, 12, 31, 0, 0, 0)
        launched_at = datetime.datetime(2014, 12, 30, 1, 0, 0)
        with self.assertRaises(pipeline_handler.UsageException) as e:
            self.handler._confirm_launched_at([],
                                              {'state': 'active',
                                               'audit_period_beginning': apb,
                                               'audit_period_ending': ape,
                                               'launched_at': launched_at})
            self.assertEquals("U8", e.code)

    def test_process_block_exists(self):
        exists = {'event_type':'compute.instance.exists', 'timestamp':'now',
                  'instance_id':'inst'}
        self.handler.stream_id = 123
        with mock.patch.object(self.handler, "_do_checks") as c:
            events = self.handler._process_block([], exists)
            self.assertEquals(1, len(events))
            f = events[0]
            self.assertEquals("compute.instance.exists.verified",
                              f['event_type'])
            self.assertEquals("now", f['timestamp'])
            self.assertEquals(123, f['stream_id'])
            self.assertEquals("inst", f['payload']['instance_id'])
            self.assertEquals("None", f['error'])
            self.assertIsNone(f['error_code'])

    def test_process_block_bad(self):
        exists = {'event_type': 'compute.instance.exists', 'timestamp':'now',
                'instance_id':'inst'}
        self.handler.stream_id = 123
        with mock.patch.object(self.handler, "_do_checks") as c:
            c.side_effect = pipeline_handler.UsageException("UX", "Error")
            events = self.handler._process_block([], exists)
            self.assertEquals(1, len(events))
            f = events[0]
            self.assertEquals("compute.instance.exists.failed",
                              f['event_type'])
            self.assertEquals("now", f['timestamp'])
            self.assertEquals(123, f['stream_id'])
            self.assertEquals("inst", f['payload']['instance_id'])
            self.assertEquals("Error", f['error'])
            self.assertEquals("UX", f['error_code'])

    def test_process_block_warnings(self):
        self.handler.warnings = ['one', 'two']
        exists = {'event_type': 'compute.instance.exists',
                  'timestamp':'now', 'instance_id':'inst'}
        self.handler.stream_id = 123
        with mock.patch.object(self.handler, "_do_checks") as c:
            events = self.handler._process_block([], exists)
            self.assertEquals(2, len(events))
            self.assertEquals("compute.instance.exists.warnings",
                              events[0]['event_type'])
            self.assertEquals("compute.instance.exists.verified",
                              events[1]['event_type'])

    @mock.patch.object(pipeline_handler.UsageHandler, '_confirm_launched_at')
    @mock.patch.object(pipeline_handler.UsageHandler, '_get_core_fields')
    @mock.patch.object(pipeline_handler.UsageHandler, '_verify_fields')
    @mock.patch.object(pipeline_handler.UsageHandler, '_is_non_EOD_exists')
    @mock.patch.object(pipeline_handler.UsageHandler, '_find_deleted_events')
    @mock.patch.object(pipeline_handler.UsageHandler, '_confirm_delete')
    def test_do_check_no_interesting_EOD_exists(self, cd, fde, inee, vf,
                                                gcf, cla):
        block = []
        exists = {'event_type': 'compute.instance.exists',
                  'message_id': 2}
        inee.return_value = False
        self.handler._do_checks(block, exists)
        self.assertTrue(len(self.handler.warnings) == 0)
        self.assertFalse(vf.called)

    @mock.patch.object(pipeline_handler.UsageHandler, '_confirm_launched_at')
    @mock.patch.object(pipeline_handler.UsageHandler, '_get_core_fields')
    @mock.patch.object(pipeline_handler.UsageHandler, '_verify_fields')
    @mock.patch.object(pipeline_handler.UsageHandler, '_is_non_EOD_exists')
    @mock.patch.object(pipeline_handler.UsageHandler, '_find_deleted_events')
    @mock.patch.object(pipeline_handler.UsageHandler, '_confirm_delete')
    def test_do_check_no_interesting_non_EOD(self, cd, fde, inee, vf,
                                             gcf, cla):
        block = []
        exists = {'event_type': 'compute.instance.exists',
                  'message_id': 2}
        inee.return_value = True
        self.handler._do_checks(block, exists)
        self.assertTrue(len(self.handler.warnings) == 1)
        self.assertFalse(vf.called)

    @mock.patch.object(pipeline_handler.UsageHandler, '_confirm_launched_at')
    @mock.patch.object(pipeline_handler.UsageHandler, '_get_core_fields')
    @mock.patch.object(pipeline_handler.UsageHandler, '_verify_fields')
    @mock.patch.object(pipeline_handler.UsageHandler, '_is_non_EOD_exists')
    @mock.patch.object(pipeline_handler.UsageHandler, '_find_deleted_events')
    @mock.patch.object(pipeline_handler.UsageHandler, '_confirm_delete')
    def test_do_check_interesting(self, cd, fde, inee, vf, gcf, cla):
        block = [{'event_type': 'compute.instance.rebuild.start',
                  'message_id': 1}]
        exists = {'event_type': 'compute.instance.exists',
                  'message_id': 2}
        self.handler._do_checks(block, exists)
        self.assertTrue(len(self.handler.warnings) == 0)
        self.assertTrue(vf.called)

    def test_handle_events_no_data(self):
        env = {'stream_id': 123}
        events = self.handler.handle_events([], env)
        self.assertEquals(0, len(events))

    def test_handle_events_no_exists(self):
        env = {'stream_id': 123}
        raw = [{'event_type': 'foo'}]
        events = self.handler.handle_events(raw, env)
        self.assertEquals(1, len(events))
        notifications = env['usage_notifications']
        self.assertEquals(1, len(notifications))
        self.assertEquals("compute.instance.exists.failed",
                          notifications[0]['event_type'])

    @mock.patch.object(pipeline_handler.UsageHandler, '_process_block')
    def test_handle_events_exists(self, pb):
        env = {'stream_id': 123}
        raw = [{'event_type': 'foo'},
               {'event_type': 'compute.instance.exists'}]
        events = self.handler.handle_events(raw, env)
        self.assertEquals(2, len(events))
        self.assertTrue(pb.called)

    @mock.patch.object(pipeline_handler.UsageHandler, '_process_block')
    def test_handle_events_dangling(self, pb):
        env = {'stream_id': 123}
        raw = [{'event_type': 'foo'},
               {'event_type': 'compute.instance.exists'},
               {'event_type': 'foo'},
               ]
        events = self.handler.handle_events(raw, env)
        self.assertEquals(3, len(events))
        notifications = env['usage_notifications']
        self.assertEquals(1, len(notifications))
        self.assertEquals("compute.instance.exists.failed",
                          notifications[0]['event_type'])
        self.assertTrue(pb.called)



import datetime
import logging
from stackdistiller import distiller, condenser
import simport

from winchester.config import ConfigManager, ConfigSection, ConfigItem
from winchester import debugging
from winchester.db import DBInterface, DuplicateError
from winchester.definition import TriggerDefinition
from winchester import time_sync as ts

logger = logging.getLogger(__name__)


class EventCondenser(condenser.CondenserBase):

    def __init__(self, dbi):
        self.dbi = dbi
        self.clear()

    def clear(self):
        self.traits = dict()
        self.event_type = None
        self.message_id = None
        self.timestamp = None

    def add_trait(self, name, trait_type, value):
        if isinstance(value, datetime.datetime):
            value = self._fix_time(value)
        self.traits[name] = value

    def add_envelope_info(self, event_type, message_id, when):
        self.event_type = event_type
        self.message_id = message_id
        self.timestamp = when

    def get_event(self):
        event = self.traits.copy()
        event['message_id'] = self.message_id
        event['timestamp'] = self._fix_time(self.timestamp)
        event['event_type'] = self.event_type
        return event

    def _fix_time(self, dt):
        """Stackdistiller converts all times to utc.
        We store timestamps as utc datetime. However, the explicit
        UTC timezone on incoming datetimes causes comparison issues
        deep in sqlalchemy. We fix this by converting all datetimes
        to naive utc timestamps"""
        if dt.tzinfo is not None:
            dt = dt.replace(tzinfo=None)
        return dt

    def validate(self):
        if self.event_type is None:
            return False
        if self.message_id is None:
            return False
        if self.timestamp is None:
            return False
        if not self.traits:
            return False
        return True


class TriggerManager(object):

    @classmethod
    def config_description(cls):
        return dict(config_path=ConfigItem(
                            help="Path(s) to find additional config files",
                                 multiple=True, default='.'),
                    distiller_config=ConfigItem(required=True,
                                       help="Name of distiller config file "
                                       "describing what to extract from the "
                                       "notifications"),
                    distiller_trait_plugins=ConfigItem(
                        help="dictionary of trait plugins to load "
                             "for stackdistiller. Classes specified with "
                             "simport syntax. See stackdistiller and "
                             "simport docs for more info", default=dict()),
                    time_sync_endpoint=ConfigItem(
                        help="URL of time sync service for use with"
                             " replying old events.",
                             default=None),
                    catch_all_notifications=ConfigItem(
                        help="Store basic info for all notifications,"
                             " even if not listed in distiller config",
                             default=False),
                    statistics_period=ConfigItem(
                        help="Emit stats on event counts, etc every "
                             "this many seconds", default=10),
                    database=ConfigSection(help="Database connection info.",
                            config_description=DBInterface.config_description()),
                    trigger_definitions=ConfigItem(required=True,
                               help="Name of trigger definitions file "
                               "defining trigger conditions and what events to "
                               "process for each stream"),
               )

    def __init__(self, config, db=None, stackdistiller=None, trigger_defs=None,
                 time_sync=None):
        config = ConfigManager.wrap(config, self.config_description())
        self.config = config
        self.debug_manager = debugging.DebugManager()
        config.check_config()
        config.add_config_path(*config['config_path'])
        if time_sync is None:
            time_sync = ts.TimeSync()
        self.time_sync = time_sync

        if db is not None:
            self.db = db
        else:
            self.db = DBInterface(config['database'])
        if stackdistiller is not None:
            self.distiller = stackdistiller
        else:
            dist_config = config.load_file(config['distiller_config'])
            plugmap = self._load_plugins(config['distiller_trait_plugins'],
                                         distiller.DEFAULT_PLUGINMAP)
            self.distiller = distiller.Distiller(dist_config,
                                                 trait_plugin_map=plugmap,
                                                 catchall=config['catch_all_notifications'])
        if trigger_defs is not None:
            self.trigger_definitions = trigger_defs
            for t in self.trigger_definitions:
                t.set_debugger(self.debug_manager)
        else:
            defs = config.load_file(config['trigger_definitions'])
            self.trigger_definitions = [TriggerDefinition(conf,
                                        self.debug_manager)
                                            for conf in defs]
        self.saved_events = 0
        self.received = 0
        self.last_status = self.current_time()

    @classmethod
    def _load_plugins(cls, plug_map, defaults=None):
        plugins = dict()
        if defaults is not None:
            plugins.update(defaults)
        for name, cls_string in plug_map.items():
            try:
                plugins[name] = simport.load(cls_string)
            except simport.ImportFailed as e:
                log.error("Could not load plugin %s: Import failed. %s" % (
                          name, e))
            except (simport.MissingMethodOrFunction,
                    simport.MissingModule,
                    simport.BadDirectory) as e:
                log.error("Could not load plugin %s: Not found. %s" % (
                          name, e))
        return plugins

    def current_time(self):
        # here so it's easily overridden.
        return self.time_sync.current_time()

    def save_event(self, event):
        traits = event.copy()
        try:
            message_id = traits.pop('message_id')
            timestamp = traits.pop('timestamp')
            event_type = traits.pop('event_type')
        except KeyError as e:
            logger.warning("Received invalid event: %s" % e)
            return False
        try:
            self.db.create_event(message_id, event_type,
                                 timestamp, traits)
            self.saved_events += 1
            return True
        except DuplicateError:
            logger.info("Received duplicate event %s, Ignoring." % message_id)
        return False

    def convert_notification(self, notification_body):
        cond = EventCondenser(self.db)
        cond.clear()
        self.received += 1
        if self.distiller.to_event(notification_body, cond):
            if cond.validate():
                return cond.get_event()
            else:
                logger.warning("Received invalid event")
        else:
            event_type = notification_body.get('event_type', '**no event_type**')
            message_id = notification_body.get('message_id', '**no id**')
            logger.info("Dropping unconverted %s notification %s" % (event_type, message_id))
        return None

    def _log_statistics(self):
        logger.info("Received %s notifications. Saved %s events." % (
                    self.received, self.saved_events))
        self.received = 0
        self.saved_events = 0
        self.last_status = self.current_time()

        self.debug_manager.dump_debuggers()

    def _add_or_create_stream(self, trigger_def, event, dist_traits):
        stream = self.db.get_active_stream(trigger_def.name, dist_traits, self.current_time())
        if stream is None:
            trigger_def.debugger.bump_counter("New stream")
            stream = self.db.create_stream(trigger_def.name, event, dist_traits,
                                           trigger_def.expiration)
            logger.debug("Created New stream %s for %s: distinguished by %s" % (
                          stream.id, trigger_def.name, str(dist_traits)))
        else:
            self.db.add_event_stream(stream, event, trigger_def.expiration)
        return stream

    def _ready_to_fire(self, stream, trigger_def):
        timestamp = trigger_def.get_fire_timestamp(self.current_time())
        self.db.stream_ready_to_fire(stream, timestamp)
        trigger_def.debugger.bump_counter("Ready to fire")
        logger.debug("Stream %s ready to fire at %s" % (
                      stream.id, timestamp))

    def add_event(self, event):
        if self.save_event(event):
            for trigger_def in self.trigger_definitions:
                matched_criteria = trigger_def.match(event)
                if matched_criteria:
                    dist_traits = trigger_def.get_distinguishing_traits(
                                                    event, matched_criteria)
                    stream = self._add_or_create_stream(trigger_def, event,
                                                        dist_traits)
                    trigger_def.debugger.bump_counter("Added events")
                    if stream.fire_timestamp is None:
                        if trigger_def.should_fire(self.db.get_stream_events(
                                                   stream)):
                            self._ready_to_fire(stream, trigger_def)

    def add_notification(self, notification_body):
        event = self.convert_notification(notification_body)
        if event:
            self.add_event(event)



import abc
import datetime
import logging
import six
import uuid

from notabene import kombu_driver as driver


logger = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class PipelineHandlerBase(object):
    """Base class for Pipeline handlers.

       Pipeline handlers perform the actual processing on a set of events
       captured by a stream. The handlers are chained together, each handler
       in a pipeline is called in order, and receives the output of the previous
       handler.

       Once all of the handlers in a pipeline have successfully processed the
       events (with .handle_events() ), each handler's .commit() method will be
       called. If any handler in the chain raises an exception, processing of
       events will stop, and each handler's .rollback() method will be called."""

    def __init__(self, **kw):
       """Setup the pipeline handler.

          A new instance of each handler for a pipeline is used for each
          stream (set of events) processed.

          :param kw: The parameters listed in the pipeline config file for
                     this handler (if any).
        """

    @abc.abstractmethod
    def handle_events(self, events, env):
        """ This method handles the actual event processing.

            This method receives a list of events and should return a list of
            events as well. The return value of this method will be passed to
            the next handler's .handle_events() method. Handlers can add new
            events simply by adding them to the list they return. New events
            (those with unrecognized message_id's), will be saved to the
            database if all handlers in this pipeline complete successfully.
            Likewise, handlers can omit events from the list they return to
            act as a filter for downstream handlers.

            Care should be taken to avoid any operation with side-effects in
            this method. Pipelines can be re-tried if a handler throws an
            error. If you need to perform such operations, such as interacting
            with an external system, save the needed information in an instance
            variable, and perform the operation in the .commit() method.

            :param events: A list of events.
            :param env:  Just a dictionary, it's passed to each handler, and
                         can act as a shared scratchpad.

            :returns: A list of events.
        """

    @abc.abstractmethod
    def commit(self):
        """ Called when each handler in this pipeline has successfully
            completed.

            If you have operations with side effects, preform them here.
            Exceptions raised here will be logged, but otherwise ignored.
        """

    @abc.abstractmethod
    def rollback(self):
        """ Called if there is an error for any handler while processing a list
        of events.

        If you need to perform some kind of cleanup, do it here.
        Exceptions raised here will be logged, but otherwise ignored.
        """


class LoggingHandler(PipelineHandlerBase):

    def handle_events(self, events, env):
        emsg = ', '.join("%s: %s" % (event['event_type'], event['message_id'])
                        for event in events)
        logger.info("Received %s events: \n%s" % (len(events), emsg))
        return events

    def commit(self):
        pass

    def rollback(self):
        pass


class NotabeneException(Exception):
    pass


class ConnectionManager(object):
    def __init__(self):
        # {connection_properties:
        #   {exchange_properties: (connection, exchange)}}
        self.pool = {}

    def _extract_params(self, kw):
        host = kw.get('host', 'localhost')
        user = kw.get('user', 'guest')
        password = kw.get('password', 'guest')
        port = kw.get('port', 5672)
        vhost = kw.get('vhost', '/')
        library = kw.get('library', 'librabbitmq')
        exchange_name = kw.get('exchange')
        exchange_type = kw.get('exchange_type', 'topic')

        if exchange_name is None:
            raise NotabeneException("No 'exchange' name provided")

        connection_dict = {'host': host, 'port': port,
                           'user': user, 'password': password,
                           'library': library, 'vhost': vhost}
        connection_tuple = tuple(sorted(connection_dict.items()))

        exchange_dict = {'exchange_name': exchange_name,
                         'exchange_type': exchange_type}
        exchange_tuple = tuple(sorted(exchange_dict.items()))

        return (connection_dict, connection_tuple,
                exchange_dict, exchange_tuple)

    def get_connection(self, properties, queue_name):
        connection_dict, connection_tuple, \
        exchange_dict, exchange_tuple = self._extract_params(properties)
        connection_info = self.pool.get(connection_tuple)
        if connection_info is None:
            connection = driver.create_connection(connection_dict['host'],
                                                  connection_dict['port'],
                                                  connection_dict['user'],
                                                  connection_dict['password'],
                                                  connection_dict['library'],
                                                  connection_dict['vhost'])
            connection_info = (connection, {})
            self.pool[connection_tuple] = connection_info
        connection, exchange_pool = connection_info
        exchange = exchange_pool.get(exchange_tuple)
        if exchange is None:
            exchange = driver.create_exchange(exchange_dict['exchange_name'],
                                              exchange_dict['exchange_type'])
            exchange_pool[exchange_tuple] = exchange

            # Make sure the queue exists so we don't lose events.
            queue = driver.create_queue(queue_name, exchange, queue_name,
                                        channel=connection.channel())
            queue.declare()

        return (connection, exchange)


# Global ConnectionManager. Shared by all Handlers.
connection_manager = ConnectionManager()


class NotabeneHandler(PipelineHandlerBase):
    # Handlers are created per stream, so we have to be smart about
    # things like connections to databases and queues.
    # We don't want to create too many connections, and we have to
    # remember that stream processing has to occur quickly, so
    # we want to avoid round-trips where possible.
    def __init__(self, **kw):
        super(NotabeneHandler, self).__init__(**kw)
        global connection_manager

        self.queue_name = kw.get('queue_name')
        if self.queue_name is None:
            raise NotabeneException("No 'queue_name' provided")
        self.connection, self.exchange = connection_manager.get_connection(
                                                        kw, self.queue_name)

        self.env_keys = kw.get('env_keys', [])

    def handle_events(self, events, env):
        keys = [key for key in self.env_keys]
        self.pending_notifications = []
        for key in keys:
            self.pending_notifications.extend(env.get(key, []))
        return events

    def commit(self):
        for notification in self.pending_notifications:
            logger.info("Publishing '%s' to '%s' with routing_key '%s'" %
                            (notification['event_type'], self.exchange,
                             self.queue_name))
            try:
                driver.send_notification(notification, self.queue_name,
                                         self.connection, self.exchange)
            except Exception as e:
                logger.exception(e)

    def rollback(self):
        pass


class UsageException(Exception):
    def __init__(self, code, message):
        super(UsageException, self).__init__(message)
        self.code = code


class UsageHandler(PipelineHandlerBase):
    def __init__(self, **kw):
        super(UsageHandler, self).__init__(**kw)
        self.warnings = []

    def _get_audit_period(self, event):
        apb = event.get('audit_period_beginning')
        ape = event.get('audit_period_ending')
        return apb, ape

    def _is_exists(self, event):
        return event['event_type'] == 'compute.instance.exists'

    def _is_non_EOD_exists(self, event):
        # For non-EOD .exists, we just check that the APB and APE are
        # not 24hrs apart. We could check that it's not midnight, but
        # could be possible (though unlikely).
        # And, if we don't find any extras, don't error out ...
        # we'll do that later.
        apb, ape = self._get_audit_period(event)
        return (self._is_exists(event) and apb and ape
            and ape.date() != (apb.date() + datetime.timedelta(days=1)))

    def _is_EOD_exists(self, event):
        # We could have several .exists records, but only the
        # end-of-day .exists will have audit_period_* time of
        # 00:00:00 and be 24hrs apart.
        apb, ape = self._get_audit_period(event)
        return (self._is_exists(event) and apb and ape
            and apb.time() == datetime.time(0, 0, 0)
            and ape.time() == datetime.time(0, 0, 0)
            and ape.date() == (apb.date() + datetime.timedelta(days=1)))

    def _extract_launched_at(self, exists):
        if not exists.get('launched_at'):
            raise UsageException("U1", ".exists has no launched_at value.")
        return exists['launched_at']

    def _extract_interesting_events(self, events, interesting):
        return [event for event in events
                        if event['event_type'] in interesting]

    def _find_deleted_events(self, events):
        interesting = ['compute.instance.delete.end']
        return self._extract_interesting_events(events, interesting)

    def _verify_fields(self, this, that, fields):
        for field in fields:
            if field not in this and field not in that:
                continue
            if this[field] != that[field]:
                raise UsageException("U2",
                                "Conflicting '%s' values ('%s' != '%s')"
                                % (field, this[field], that[field]))

    def _confirm_delete(self, exists, delete_events, fields):
        deleted_at = exists.get('deleted_at')
        state = exists.get('state')
        apb, ape = self._get_audit_period(exists)

        if not deleted_at and delete_events:
            raise UsageException("U6", ".deleted events found but .exists "
                                "has no deleted_at value.")

        if deleted_at and state != "deleted":
            raise UsageException("U3", ".exists state not 'deleted' but "
                             ".exists deleted_at is set.")

        if deleted_at and not delete_events:
            # We've already confirmed it's in the "deleted" state.
            launched_at = exists.get('launched_at')
            if deleted_at < launched_at:
                raise UsageException("U4",
                              ".exists deleted_at < .exists launched_at.")

            # Is the deleted_at within this audit period?
            if (apb and ape and deleted_at >= apb and deleted_at <= ape):
                raise UsageException("U5", ".exists deleted_at in audit "
                    "period, but no matching .delete event found.")

        if len(delete_events) > 1:
            raise UsageException("U7", "Multiple .delete.end events")

        if delete_events:
            self._verify_fields(exists, delete_events[0],  fields)

    def _confirm_launched_at(self, block, exists):
        if exists.get('state') != 'active':
            return

        apb, ape = self._get_audit_period(exists)

        # Does launched_at have a value within this audit period?
        # If so, we should have a related event. Otherwise, this
        # instance was created previously.
        launched_at = self._extract_launched_at(exists)
        if (apb and ape and
            launched_at >= apb and launched_at <= ape and
            len(block) == 0):
                raise UsageException("U8", ".exists launched_at in audit "
                    "period, but no related events found.")

        # TODO(sandy): Confirm the events we got set launched_at
        # properly.

    def _get_core_fields(self):
        """Broken out so derived classes can define their
           own trait list."""
        return ['launched_at', 'instance_flavor_id', 'tenant_id',
                'os_architecture', 'os_version', 'os_distro']

    def _do_checks(self, block, exists):
        interesting = ['compute.instance.rebuild.start',
                       'compute.instance.resize.prep.start',
                       'compute.instance.rescue.start']

        self._confirm_launched_at(block, exists)

        fields = self._get_core_fields()
        last_interesting = None
        for event in block:
            if event['event_type'] in interesting:
                last_interesting = event
        if last_interesting:
            self._verify_fields(last_interesting, exists, fields)
        elif self._is_non_EOD_exists(exists):
            self.warnings.append("Non-EOD .exists found "
                                 "(msg_id: %s) "
                                 "with no operation event." %
                                 exists['message_id'])

        deleted = self._find_deleted_events(block)
        delete_fields = ['launched_at', 'deleted_at']
        self._confirm_delete(exists, deleted, delete_fields)

    def _base_notification(self, exists):
        apb, ape = self._get_audit_period(exists)
        return {
            'payload': {
              'audit_period_beginning': str(apb),
              'audit_period_ending': str(ape),
              'launched_at': str(exists.get('launched_at', '')),
              'deleted_at': str(exists.get('deleted_at', '')),
              'instance_id': exists.get('instance_id', ''),
              'tenant_id': exists.get('tenant_id', ''),
              'display_name': exists.get('display_name', ''),
              'instance_type': exists.get('instance_flavor', ''),
              'instance_flavor_id': exists.get('instance_flavor_id', ''),
              'state': exists.get('state', ''),
              'state_description': exists.get('state_description', ''),
              'bandwidth': {'public': {
                              'bw_in': exists.get('bandwidth_in', 0),
                              'bw_out': exists.get('bandwidth_out', 0)}},
              'image_meta': {
                'org.openstack__1__architecture':
                                            exists.get('os_architecture', ''),
                'org.openstack__1__os_version': exists.get('os_version', ''),
                'org.openstack__1__os_distro': exists.get('os_distro', ''),
                'org.rackspace__1__options': exists.get('rax_options', '0')
              }},
            'original_message_id': exists.get('message_id', '')}

    def _process_block(self, block, exists):
        error = None
        try:
            self._do_checks(block, exists)
            event_type = "compute.instance.exists.verified"
        except UsageException as e:
            error = e
            event_type = "compute.instance.exists.failed"
            logger.warn("Stream %s UsageException: (%s) %s" %
                                            (self.stream_id, e.code, e))
            apb, ape = self._get_audit_period(exists)
            logger.warn("Stream %s deleted_at: %s, launched_at: %s, "
                        "state: %s, APB: %s, APE: %s, #events: %s" %
                        (self.stream_id, exists.get("deleted_at"),
                         exists.get("launched_at"), exists.get("state"),
                         apb, ape, len(block)))

        if len(block) > 1:
            logger.warn("%s - events (stream: %s)"
                                    % (event_type, self.stream_id))
            for event in block:
                logger.warn("^Event: %s - %s" %
                                    (event['timestamp'], event['event_type']))

        events = []
        # We could have warnings, but a valid event list.
        if self.warnings:
            instance_id = exists.get('instance_id', 'n/a')
            warning_event = {'event_type': 'compute.instance.exists.warnings',
                             'publisher_id': 'stv3',
                             'message_id': str(uuid.uuid4()),
                             'timestamp': exists.get('timestamp',
                                                  datetime.datetime.utcnow()),
                             'stream_id': int(self.stream_id),
                             'instance_id': exists.get('instance_id'),
                             'warnings': self.warnings}
            events.append(warning_event)

        new_event = self._base_notification(exists)
        new_event.update({'event_type': event_type,
                          'message_id': str(uuid.uuid4()),
                          'publisher_id': 'stv3',
                          'timestamp': exists.get('timestamp',
                                             datetime.datetime.utcnow()),
                          'stream_id': int(self.stream_id),
                          'error': str(error),
                          'error_code': error and error.code})
        events.append(new_event)
        return events

    def handle_events(self, events, env):
        self.env = env
        self.stream_id = env['stream_id']
        self.warnings = []

        new_events = []
        block = []
        for event in events:
            if self._is_exists(event):
                new_events.extend(self._process_block(block, event))
                block = []
            else:
                block.append(event)

        # Final block should be empty.
        if block:
            new_event = {'event_type': "compute.instance.exists.failed",
                         'message_id': str(uuid.uuid4()),
                         'timestamp': block[0].get('timestamp',
                                                   datetime.datetime.utcnow()),
                         'stream_id': int(self.stream_id),
                         'instance_id': block[0].get('instance_id'),
                         'error': "Notifications, but no .exists "
                                  "notification found.",
                         'error_code': "U0"
                        }
            new_events.append(new_event)

        env['usage_notifications'] = new_events
        return events

    def commit(self):
        pass

    def rollback(self):
        pass



import argparse
import alembic
import logging

from winchester.db.alembic_command import AlembicCommandLine


logger = logging.getLogger(__name__)


class DBAdminCommandLine(AlembicCommandLine):
    description = "Winchester DB admin commandline tool."

    def add_options(self, parser):
        parser.add_argument('--config', '-c',
                            default='winchester.yaml',
                            type=str,
                            help='The name of the winchester config file')


    def get_config(self, options):
        alembic_cfg = alembic.config.Config()
        alembic_cfg.set_main_option("winchester_config", options.config)
        alembic_cfg.set_main_option("script_location", "winchester.db:migrations")
        return alembic_cfg


def main():
    cmd = DBAdminCommandLine(allowed_commands=['upgrade', 'downgrade',
                                               'current', 'history', 'stamp'])
    cmd.main()


if __name__ == '__main__':
    main()



import logging
import collections
import datetime
import six
import timex
import fnmatch


logger = logging.getLogger(__name__)


class DefinitionError(Exception):
    pass


def filter_event_timestamps(event):
    return dict((trait, value) for trait, value in event.items()
                if isinstance(value, datetime.datetime))


class Criterion(object):

    @classmethod
    def get_from_expression(cls, expression, trait_name):
        if isinstance(expression, collections.Mapping):
            if len(expression) != 1:
                raise DefinitionError("Only exactly one type of match is "
                                      "allowed per criterion expression")
            ctype = expression.keys()[0]
            expr = expression[ctype]
            if ctype == 'int':
                return NumericCriterion(expr, trait_name)
            elif ctype =='float':
                return FloatCriterion(expr, trait_name)
            elif ctype == 'datetime':
                return TimeCriterion(expr, trait_name)
            elif ctype == 'string' or ctype == 'text':
                return Criterion(expr, trait_name)
        else:
            # A constant. -mdragon
            return Criterion(expression, trait_name)

    def __init__(self, expr, trait_name):
        self.trait_name = trait_name
        #match a constant
        self.op = '='
        self.value = expr

    def match(self, event, debug_group):
        if self.trait_name not in event:
            return debug_group.mismatch("not %s" % self.trait_name)
        value = event[self.trait_name]
        if self.op == '=':
            return debug_group.check(value == self.value, "== failed")
        elif self.op == '>':
            return debug_group.check(value > self.value, "> failed")
        elif self.op == '<':
            return debug_group.check(value < self.value, "< failed")
        return debug_group.mismatch("Criterion match() fall-thru")


class NumericCriterion(Criterion):

    def __init__(self, expr, trait_name):
        self.trait_name = trait_name
        if not isinstance(expr, six.string_types):
            self.op = '='
            self.value = expr
        else:
            expr = expr.strip().split(None, 1)
            if len(expr) == 2:
                self.op = expr[0]
                value = expr[1].strip()
            elif len(expr) == 1:
                self.op = '='
                value = expr[0]
            else:
                raise DefinitionError('Invalid numeric criterion.')
            try:
                self.value = self._convert(value)
            except ValueError:
                raise DefinitionError('Invalid numeric criterion.')

    def _convert(self, value):
        return int(value)


class FloatCriterion(NumericCriterion):

    def _convert(self, value):
        return float(value)


class TimeCriterion(Criterion):

    def __init__(self, expression, trait_name):
        self.trait_name = trait_name
        self.time_expr = timex.parse(expression)

    def match(self, event, debug_group):
        if self.trait_name not in event:
            return debug_group.mismatch("Time: not '%s'" % self.trait_name)
        value = event[self.trait_name]
        try:
            timerange = self.time_expr(**filter_event_timestamps(event))
        except timex.TimexExpressionError:
            # the event doesn't contain a trait referenced in the expression.
            return debug_group.mismatch("Time: no referenced trait")
        return debug_group.check(value in timerange, "Not in timerange")


class Criteria(object):
    def __init__(self, config):
        self.included_types = []
        self.excluded_types = []
        if 'event_type' in config:
            event_types = config['event_type']
            if isinstance(event_types, six.string_types):
                event_types = [event_types]
            for t in event_types:
                if t.startswith('!'):
                    self.excluded_types.append(t[1:])
                else:
                    self.included_types.append(t)
        else:
            self.included_types.append('*')
        if self.excluded_types and not self.included_types:
            self.included_types.append('*')
        if 'number' in config:
            self.number = config['number']
        else:
            self.number = 1
        if 'timestamp' in config:
            self.timestamp = timex.parse(config['timestamp'])
        else:
            self.timestamp = None
        self.map_distinguished_by = dict()
        if 'map_distinguished_by' in config:
            self.map_distinguished_by = config['map_distinguished_by']
        self.traits = dict()
        if 'traits' in config:
            for trait, criterion in config['traits'].items():
                self.traits[trait] = Criterion.get_from_expression(criterion, trait)

    def included_type(self, event_type):
        return any(fnmatch.fnmatch(event_type, t) for t in self.included_types)

    def excluded_type(self, event_type):
        return any(fnmatch.fnmatch(event_type, t) for t in self.excluded_types)

    def match_type(self, event_type):
        return (self.included_type(event_type)
                and not self.excluded_type(event_type))

    def match(self, event, debug_group):
        if not self.match_type(event['event_type']):
            return debug_group.mismatch("Wrong event type")
        if self.timestamp:
            try:
                t = self.timestamp(**filter_event_timestamps(event))
            except timex.TimexExpressionError:
                # the event doesn't contain a trait referenced in the expression.
                return debug_group.mismatch("No timestamp trait")
            if event['timestamp'] not in t:
                return debug_group.mismatch("Not time yet.")
        if not self.traits:
            return debug_group.match()
        return all(criterion.match(event, debug_group) for
                       criterion in self.traits.values())


class TriggerDefinition(object):

    def __init__(self, config, debug_manager):
        if 'name' not in config:
            raise DefinitionError("Required field in trigger definition not "
                                  "specified 'name'")
        self.name = config['name']
        self.debug_level = int(config.get('debug_level', 0))
        self.distinguished_by = config.get('distinguished_by', [])
        for dt in self.distinguished_by:
            if isinstance(dt, collections.Mapping):
                if len(dt) > 1:
                    raise DefinitionError("Invalid distinguising expression "
                        "%s. Only one trait allowed in an expression" % str(dt))
        self.fire_delay = config.get('fire_delay', 0)
        if 'expiration' not in config:
            raise DefinitionError("Required field in trigger definition not "
                                  "specified 'expiration'")
        self.expiration = timex.parse(config['expiration'])
        self.fire_pipeline = config.get('fire_pipeline')
        self.expire_pipeline = config.get('expire_pipeline')
        if not self.fire_pipeline and not self.expire_pipeline:
            raise DefinitionError("At least one of: 'fire_pipeline' or "
                                  "'expire_pipeline' must be specified in a "
                                  "trigger definition.")
        if 'fire_criteria' not in config:
            raise DefinitionError("Required criteria in trigger definition not "
                                  "specified 'fire_criteria'")
        self.fire_criteria = [Criteria(c) for c in config['fire_criteria']]
        if 'match_criteria' not in config:
            raise DefinitionError("Required criteria in trigger definition not "
                                  "specified 'match_criteria'")
        self.match_criteria = [Criteria(c) for c in config['match_criteria']]
        self.load_criteria = []
        if 'load_criteria' in config:
            self.load_criteria = [Criteria(c) for c in config['load_criteria']]
        if debug_manager:
            self.set_debugger(debug_manager)

    def set_debugger(self, debug_manager):
        self.debugger = debug_manager.get_debugger(self)

    def match(self, event):
        # all distinguishing traits must exist to match.
        group = self.debugger.get_group("Match")
        for dt in self.distinguished_by:
            if isinstance(dt, collections.Mapping):
                trait_name = dt.keys()[0]
            else:
                trait_name = dt
            if trait_name not in event:
                group.mismatch("not '%s'" % trait_name)
                return None

        for criteria in self.match_criteria:
            if criteria.match(event, group):
                group.match()
                return criteria

        group.mismatch("No matching criteria")
        return None

    def get_distinguishing_traits(self, event, matching_criteria):
        dist_traits = dict()
        for dt in self.distinguished_by:
            d_expr = None
            if isinstance(dt, collections.Mapping):
                trait_name = dt.keys()[0]
                d_expr = timex.parse(dt[trait_name])
            else:
                trait_name = dt
            event_trait_name = matching_criteria.map_distinguished_by.get(trait_name, trait_name)
            if d_expr is not None:
                dist_traits[trait_name] = d_expr(timestamp=event[event_trait_name])
            else:
                dist_traits[trait_name] = event[event_trait_name]
        return dist_traits

    def get_fire_timestamp(self, timestamp):
        return timestamp + datetime.timedelta(seconds=self.fire_delay)

    def should_fire(self, events):
        group = self.debugger.get_group("Fire")
        for criteria in self.fire_criteria:
            matches = 0
            for event in events:
                if criteria.match(event, group):
                    matches += 1
                    if matches >= criteria.number:
                        break
            if matches < criteria.number:
                return group.mismatch("Not enough matching criteria")
        return group.match()



import unittest2 as unittest

import datetime
import mock

from winchester import pipeline_handler


class TestConnectionManager(unittest.TestCase):
    def setUp(self):
        super(TestConnectionManager, self).setUp()
        self.mgr = pipeline_handler.ConnectionManager()

    def test_extract_params(self):
        with self.assertRaises(pipeline_handler.NotabeneException):
            self.mgr._extract_params({})

        cd, ct, ed, et = self.mgr._extract_params({'exchange': 'my_exchange'})

        self.assertEquals(cd, {'host': 'localhost',
                               'port': 5672,
                               'user': 'guest',
                               'password': 'guest',
                               'library': 'librabbitmq',
                               'vhost': '/'})

        self.assertEquals(ct, (('host', 'localhost'),
                               ('library', 'librabbitmq'),
                               ('password', 'guest'),
                               ('port', 5672),
                               ('user', 'guest'),
                               ('vhost', '/')))

        self.assertEquals(ed, {'exchange_name': 'my_exchange',
                               'exchange_type': 'topic'})

        self.assertEquals(et, (('exchange_name', 'my_exchange'),
                               ('exchange_type', 'topic')))


        kw = {'host': 'my_host', 'user': 'my_user', 'password': 'pwd',
              'port': 123, 'vhost': 'virtual', 'library': 'my_lib',
              'exchange': 'my_exchange', 'exchange_type': 'foo'}

        cd, ct, ed, et = self.mgr._extract_params(kw)

        self.assertEquals(cd, {'host': 'my_host',
                               'port': 123,
                               'user': 'my_user',
                               'password': 'pwd',
                               'library': 'my_lib',
                               'vhost': 'virtual'})

        self.assertEquals(ct, (('host', 'my_host'),
                               ('library', 'my_lib'),
                               ('password', 'pwd'),
                               ('port', 123),
                               ('user', 'my_user'),
                               ('vhost', 'virtual')))

        self.assertEquals(ed, {'exchange_name': 'my_exchange',
                               'exchange_type': 'foo'})

        self.assertEquals(et, (('exchange_name', 'my_exchange'),
                               ('exchange_type', 'foo')))


    @mock.patch.object(pipeline_handler.ConnectionManager, '_extract_params')
    @mock.patch.object(pipeline_handler.driver, 'create_connection')
    @mock.patch.object(pipeline_handler.driver, 'create_exchange')
    @mock.patch.object(pipeline_handler.driver, 'create_queue')
    def test_get_connection(self, cq, ce, cc, ep):
        conn = {'host': 'my_host', 'user': 'my_user', 'password': 'pwd',
              'port': 123, 'vhost': 'virtual', 'library': 'my_lib'}
        conn_set = tuple(sorted(conn.items()))
        exchange = {'exchange_name': 'my_exchange', 'exchange_type': 'foo'}
        exchange_set = tuple(sorted(exchange.items()))

        ep.return_value = (conn, conn_set, exchange, exchange_set)
        connection = mock.MagicMock("the connection")
        channel = mock.MagicMock("the channel")
        connection.channel = channel
        cc.return_value = connection
        mexchange = mock.MagicMock("the exchange")
        ce.return_value = mexchange
        queue = mock.MagicMock("the queue")
        queue.declare = mock.MagicMock()
        cq.return_value = queue

        final_connection, final_exchange = self.mgr.get_connection({}, "foo")

        self.assertEquals(final_connection, connection)
        self.assertEquals(final_exchange, mexchange)
        self.assertEquals(1, queue.declare.call_count)

        # Calling again should give the same results ...
        final_connection, final_exchange = self.mgr.get_connection({}, "foo")

        self.assertEquals(final_connection, connection)
        self.assertEquals(final_exchange, mexchange)
        self.assertTrue(queue.declare.called)
        self.assertEquals(1, queue.declare.call_count)

        # Change the exchange, and we should have same connection, but new
        # exchange object.
        exchange2 = {'exchange_name': 'my_exchange2', 'exchange_type': 'foo2'}
        exchange2_set = tuple(sorted(exchange2.items()))

        ep.return_value = (conn, conn_set, exchange2, exchange2_set)
        mexchange2 = mock.MagicMock("the exchange 2")
        ce.return_value = mexchange2

        final_connection, final_exchange = self.mgr.get_connection({}, "foo")

        self.assertEquals(final_connection, connection)
        self.assertEquals(final_exchange, mexchange2)
        self.assertEquals(2, queue.declare.call_count)

        # Change the connection, and we should have a new connection and new
        # exchange object.
        conn2 = {'host': 'my_host2', 'user': 'my_user2', 'password': 'pwd2',
              'port': 1234, 'vhost': 'virtual2', 'library': 'my_lib2'}
        conn2_set = tuple(sorted(conn2.items()))
        exchange3= {'exchange_name': 'my_exchange', 'exchange_type': 'foo'}
        exchange3_set = tuple(sorted(exchange3.items()))

        ep.return_value = (conn2, conn2_set, exchange3, exchange3_set)
        mexchange3 = mock.MagicMock("the exchange 3")
        ce.return_value = mexchange3

        connection2 = mock.MagicMock("the connection 2")
        channel2 = mock.MagicMock("the channel 2")
        connection2.channel = channel2
        cc.return_value = connection2

        final_connection, final_exchange = self.mgr.get_connection({}, "foo")

        self.assertEquals(final_connection, connection2)
        self.assertEquals(final_exchange, mexchange3)
        self.assertEquals(3, queue.declare.call_count)


class TestException(Exception):
    pass


class TestNotabeneHandler(unittest.TestCase):

    def test_constructor_no_queue(self):
        with self.assertRaises(pipeline_handler.NotabeneException) as e:
            pipeline_handler.NotabeneHandler()

    @mock.patch.object(pipeline_handler.connection_manager, 'get_connection')
    def test_constructor_queue(self, cm):
        cm.return_value = (1, 2)
        kw = {'queue_name': 'foo'}
        h = pipeline_handler.NotabeneHandler(**kw)
        self.assertIsNotNone(h.connection)
        self.assertIsNotNone(h.exchange)
        self.assertEquals(h.env_keys, [])

    @mock.patch.object(pipeline_handler.connection_manager, 'get_connection')
    def test_constructor_env_keys(self, cm):
        cm.return_value = (1, 2)
        kw = {'queue_name': 'foo', 'env_keys': ['x', 'y']}
        h = pipeline_handler.NotabeneHandler(**kw)
        self.assertIsNotNone(h.connection)
        self.assertIsNotNone(h.exchange)
        self.assertEquals(h.env_keys, ['x', 'y'])

    @mock.patch.object(pipeline_handler.connection_manager, 'get_connection')
    def test_handle_events(self, cm):
        cm.return_value = (1, 2)
        kw = {'queue_name': 'foo', 'env_keys': ['x', 'y', 'z']}
        h = pipeline_handler.NotabeneHandler(**kw)
        events = range(5)
        env = {'x': ['cat', 'dog'], 'y': ['fish']}
        ret = h.handle_events(events, env)
        self.assertEquals(ret, events)
        self.assertEquals(h.pending_notifications, ['cat', 'dog', 'fish'])

    @mock.patch.object(pipeline_handler.connection_manager, 'get_connection')
    def test_commit_good(self, cm):
        cm.return_value = (1, 2)
        kw = {'queue_name': 'foo'}
        h = pipeline_handler.NotabeneHandler(**kw)

        h.pending_notifications = [{'event_type': 'event1'},
                                   {'event_type': 'event2'}]
        with mock.patch.object(pipeline_handler.driver,
                                            'send_notification') as sn:
            h.commit()
            self.assertEquals(sn.call_count, 2)

    @mock.patch.object(pipeline_handler.connection_manager, 'get_connection')
    def test_commit(self, cm):
        cm.return_value = (1, 2)
        kw = {'queue_name': 'foo'}
        h = pipeline_handler.NotabeneHandler(**kw)

        h.pending_notifications = [{'event_type': 'event1'},
                                   {'event_type': 'event2'}]
        with mock.patch.object(pipeline_handler.driver,
                                            'send_notification') as sn:
            sn.side_effect = TestException
            with mock.patch.object(pipeline_handler.logger,
                                                    'exception') as ex:
                h.commit()
                self.assertEquals(ex.call_count, 2)
            self.assertEquals(sn.call_count, 2)
